"""
Implementation of an ImageFormat class to read SMV format image

but not - in the first instance - actually provide a full image
representation. This is simply there to set everything up for the ADSC
and Rigaku Saturn image readers which really will acquire the full image
including header information and generate the experimental model
representations.
"""

from __future__ import annotations

from boost_adaptbx.boost.python import streambuf
from scitbx.array_family import flex

from dxtbx import IncorrectFormatError
from dxtbx.ext import is_big_endian, read_uint16, read_uint16_bs
from dxtbx.format.Format import Format


class FormatSMV(Format):
    """An image reading class for SMV format images i.e. those from ADSC and
    Rigaku which start with:

    {
    HEADER_BYTES=  512;

    and contain a list of keyword-value pairs thereafter which define the
    header. The keywords etc. for these will depend on the instrument
    manufacturer and will be interpreted by subclasses of this class. Note
    also that every line is finished with a semicolon."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an SMV format image, i.e. we can
        make sense of it."""
        with FormatSMV.open_file(image_file, "rb") as fh:
            return fh.read(15) == b"{\nHEADER_BYTES="

    @staticmethod
    def get_smv_header(image_file):
        with FormatSMV.open_file(image_file, "rb") as fh:
            header_info = fh.read(45).decode("ascii", "ignore")
            header_size = int(
                header_info.split("\n")[1].split("=")[1].replace(";", "").strip()
            )
            fh.seek(0)
            header_text = fh.read(header_size).decode("ascii", "ignore")
        header_dictionary = {}

        # Check that we have the whole header, contained within { }.  Stop
        # extracting data once a record solely composed of a closing curly
        # brace is seen.  If there is no such character in header_text
        # either HEADER_BYTES caused a short read of the header or the
        # header is malformed.
        for record in header_text.split("\n"):
            if record == "}":
                break
            if "=" not in record:
                continue

            key, value = record.replace(";", "").split("=")

            header_dictionary[key.strip()] = value.strip()

        return header_size, header_dictionary

    def _get_endianic_raw_data(self, size):
        big_endian = self._header_dictionary["BYTE_ORDER"] == "big_endian"

        with self.open_file(self._image_file, "rb") as fh:
            fh.seek(self._header_size)

            if big_endian == is_big_endian():
                raw_data = read_uint16(streambuf(fh), int(size[0] * size[1]))
            else:
                raw_data = read_uint16_bs(streambuf(fh), int(size[0] * size[1]))

        # note that x and y are reversed here
        raw_data.reshape(flex.grid(size[1], size[0]))
        return raw_data

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        super().__init__(image_file, **kwargs)

    def _start(self):
        """Open the image file, read the image header, copy the key / value
        pairs into an internal dictionary self._header_dictionary along with
        the length of the header in bytes self._header_size."""

        self._header_size, self._header_dictionary = FormatSMV.get_smv_header(
            self._image_file
        )

    def get_beam_direction(self):
        return [
            float(sv) for sv in self._header_dictionary["SOURCE_VECTORS"].split()[:3]
        ]

    def get_beam_pixels(self, detector_name):
        return [
            float(bp)
            for bp in self._header_dictionary[
                "%sSPATIAL_DISTORTION_INFO" % detector_name
            ].split()[:2]
        ]

    def get_beam_polarization(self):
        polarization = [
            float(sp) for sp in self._header_dictionary["SOURCE_POLARZ"].split()
        ]
        p_fraction = polarization[0]
        p_plane = polarization[1:]
        return p_fraction, p_plane

    def get_detector_axes(self, detector_name):
        return [
            float(v)
            for v in self._header_dictionary[
                "%sDETECTOR_VECTORS" % detector_name
            ].split()
        ]

    def get_distortion(self, detector_name):
        return [
            float(sdv)
            for sdv in self._header_dictionary[
                "%sSPATIAL_DISTORTION_VECTORS" % detector_name
            ].split()
        ]

    def get_gonio_axes(self, detector_name):
        return [
            float(gv)
            for gv in self._header_dictionary["%sGONIO_VECTORS" % detector_name].split()
        ]

    def get_gonio_values(self, detector_name):
        return [
            float(gv)
            for gv in self._header_dictionary["%sGONIO_VALUES" % detector_name].split()
        ]

    def get_image_size(self, detector_name):
        return [
            int(dd)
            for dd in self._header_dictionary[
                "%sDETECTOR_DIMENSIONS" % detector_name
            ].split()
        ]

    def get_pixel_size(self, detector_name):
        return [
            float(ps)
            for ps in self._header_dictionary[
                "%sSPATIAL_DISTORTION_INFO" % detector_name
            ].split()[2:]
        ]

    def get_rotation(self):
        return [float(r) for r in self._header_dictionary["ROTATION"].split()]



from __future__ import annotations

import pickle

import pytest

from scitbx.array_family import flex

from dxtbx.format.FormatHDF5SaclaMPCCD import FormatHDF5SaclaMPCCD
from dxtbx.format.image import ImageBool
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx_masking_ext import mask_untrusted_resolution_range

pytest.importorskip("h5py")


def test_static_mask(dials_data):
    master_h5 = dials_data("image_examples") / "SACLA-MPCCD-Phase3-21528-5images.h5"
    assert FormatHDF5SaclaMPCCD.understand(master_h5)

    expts_from_filename = ExperimentListFactory.from_filenames([master_h5])
    expts_from_dict = ExperimentListFactory.from_dict(expts_from_filename.to_dict())
    for expts in (expts_from_filename, expts_from_dict):
        assert len(expts) == 5
        imageset = expts[0].imageset
        assert imageset.get_format_class() == FormatHDF5SaclaMPCCD
        mask = imageset.get_mask(0)
        assert len(mask) == 8
        assert [m.count(False) for m in mask] == [24296] * 8

    imageset.reader().nullify_format_instance()


def test_MPCCD_Phase3_21528(dials_data):
    master_h5 = dials_data("image_examples") / "SACLA-MPCCD-Phase3-21528-5images.h5"
    assert FormatHDF5SaclaMPCCD.understand(master_h5)
    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatHDF5SaclaMPCCD

    detector = imageset.get_detector()
    beam = imageset.get_beam()

    assert len(detector) == 8
    panel = detector[0]
    assert panel.get_pixel_size() == pytest.approx((0.05, 0.05))
    assert panel.get_image_size() == (512, 1024)
    assert panel.get_trusted_range() == (0, 65535)
    assert panel.get_fast_axis() == pytest.approx(
        (-0.0026929852392799875, -0.9999963739086762, 0.0)
    )
    assert panel.get_slow_axis() == pytest.approx(
        (-0.9999963739086762, 0.0026929852392799875, 0.0)
    )
    assert panel.get_thickness() == pytest.approx(0.3)
    assert panel.get_mu() == 0
    assert panel.get_material() == ""
    assert panel.get_origin() == pytest.approx(
        (-1.8203499755859376, 51.6243984375, -99.5)
    )
    assert panel.get_distance() == pytest.approx(99.5)

    assert imageset.get_goniometer() is None
    assert imageset.get_scan() is None

    assert beam.get_wavelength() == pytest.approx(1.2452863763694901)
    assert beam.get_sample_to_source_direction() == (0, 0, 1)
    imageset.reader().nullify_format_instance()


def test_MPCCD_RECONST_MODE(dials_data, monkeypatch):
    for MPCCD_RECONST_MODE in (0, 1):
        if MPCCD_RECONST_MODE:
            monkeypatch.setenv("MPCCD_RECONST_MODE", str(MPCCD_RECONST_MODE))
        else:
            monkeypatch.delenv("MPCCD_RECONST_MODE", raising=False)

        master_h5 = dials_data("image_examples") / "SACLA-MPCCD-Phase3-21528-5images.h5"
        expts = ExperimentListFactory.from_filenames([master_h5])
        imageset = expts[0].imageset
        # Horrible hack to work around format_instance caching
        imageset.reader().nullify_format_instance()
        raw_data = imageset.get_raw_data(0)
        mmm = raw_data[0].as_double().as_1d().min_max_mean()
        mmm.show()
        if MPCCD_RECONST_MODE:
            assert mmm.min == -1
            assert mmm.max == 14102
            assert mmm.mean == pytest.approx(2.93752665030144)
        else:
            assert mmm.min == 0
            assert mmm.max == 3610.0
            assert mmm.mean == pytest.approx(3.397266387939453)
    # Horrible hack to work around format_instance caching
    # This is needed to prevent an incorrect format_instance affecting
    # other tests that are run after this one.
    imageset.reader().nullify_format_instance()


def test_combine_with_user_static_mask(dials_data, tmpdir):
    master_h5 = dials_data("image_examples") / "SACLA-MPCCD-Phase3-21528-5images.h5"
    assert FormatHDF5SaclaMPCCD.understand(master_h5)

    expts_from_filename = ExperimentListFactory.from_filenames([master_h5])
    for i, expt in enumerate(expts_from_filename):
        mask = []
        for panel in expt.detector:
            m = flex.bool(flex.grid(reversed(panel.get_image_size())), True)
            mask_untrusted_resolution_range(m, expt.beam, panel, 0, 2.23)
            mask.append(m)
        mask = tuple(mask)
        # exact number of masked pixels varies with wavelength between experiments
        assert [_.count(False) for _ in mask] == pytest.approx(
            [50643, 0, 0, 48003, 48356, 0, 0, 52191], abs=1e3
        )
        mask_file = tmpdir.join("pixel_%i.mask" % i)
        with mask_file.open("wb") as f:
            pickle.dump(mask, f)
        expt.imageset.external_lookup.mask.filename = mask_file.strpath
        expt.imageset.external_lookup.mask.data = ImageBool(mask)

    expts_from_dict = ExperimentListFactory.from_dict(expts_from_filename.to_dict())
    imageset = expts_from_dict[0].imageset
    mask = imageset.get_mask(0)
    assert len(mask) == 8
    assert [_.count(False) for _ in mask] == [
        65332,
        24296,
        24296,
        63335,
        63660,
        24296,
        24296,
        66691,
    ]
    imageset.reader().nullify_format_instance()


@pytest.mark.parametrize(("clear_cache"), [False, True])
def test_HDF5_format_caching(dials_data, clear_cache):
    """
    xfail: see https://github.com/cctbx/dxtbx/issues/245
    """
    img_file = "SACLA-MPCCD-Phase3-21528-5images.h5"
    master_h5 = dials_data("image_examples") / img_file

    expts1 = ExperimentListFactory.from_filenames([master_h5])
    expts1[0].imageset.get_mask(0)
    if clear_cache:
        expts1[0].imageset.reader().nullify_format_instance()
    expts2 = ExperimentListFactory.from_filenames([master_h5])

    for e1, e2 in zip(expts1, expts2):
        assert (
            e1.imageset.get_beam().get_wavelength()
            == e2.imageset.get_beam().get_wavelength()
        )



from __future__ import annotations

import glob
import os

from dxtbx.sequence_filenames import template_string_to_glob_expr


def resolve_path(path, directory=None):
    """Resolve a file path.

    First expand any environment and user variables. Then create the absolute
    path by applying the relative path to the provided directory, if necessary.

    Args:
      path (str): The path to resolve
      directory (Optional[str]): The local path to resolve relative links

    Returns:
        str: The absolute path to the file to read if accessible, otherwise
        return the original path as provided
    """
    if not path:
        return ""
    trial_path = os.path.expanduser(os.path.expandvars(path))
    if directory and not os.path.isabs(trial_path):
        trial_path = os.path.join(directory, trial_path)
    trial_path = os.path.abspath(trial_path)
    if glob.glob(template_string_to_glob_expr(trial_path)):
        return trial_path
    else:
        return path



from __future__ import annotations

import re

import h5py

from dxtbx.format.FormatNXmxEigerFilewriter import FormatNXmxEigerFilewriter

DATA_FILE_RE = re.compile(r"data_\d{6}")


class FormatNXmxEigerFilewriterESRFID232(FormatNXmxEigerFilewriter):
    _cached_file_handle = None

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file) as handle:
            if "/entry/instrument/detector/detector_number" in handle:
                if (
                    handle["/entry/instrument/detector/detector_number"][()]
                    == b"E-18-0133"
                ):
                    return True
        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""
        super().__init__(image_file, **kwargs)

    def _start(self):
        super()._start()

    def _goniometer(self):
        return self._goniometer_factory.known_axis((0, 1, 0))



from __future__ import annotations

import math
import random

import pytest

from cctbx import crystal, sgtbx, uctbx
from cctbx.sgtbx import change_of_basis_op
from iotbx.phil import parse
from libtbx.test_utils import approx_equal
from scitbx import matrix
from scitbx.array_family import flex
from scitbx.math import euler_angles, euler_angles_as_matrix

from dxtbx.model import (
    Crystal,
    CrystalFactory,
    MosaicCrystalKabsch2010,
    MosaicCrystalSauter2014,
)
from dxtbx.model.crystal import crystal_phil_scope

from .crystal_model_old import crystal_model_old


def random_rotation():
    return euler_angles_as_matrix([random.uniform(0, 360) for i in range(3)])


def test_crystal_model_from_mosflm_matrix():
    mosflm_matrix = [
        float(v)
        for v in """ -0.00495480 -0.01491776  0.00238445
  0.01505572 -0.00661190 -0.00149401
  0.00585043  0.00438127  0.00586415
       0.000       0.000       0.000
  -0.2932645  -0.8829514   0.3665960
   0.8911171  -0.3913446  -0.2296951
   0.3462750   0.2593185   0.9015806
     57.7822     57.7822    150.0931     90.0000     90.0000     90.0000
       0.000       0.000       0.000""".split()
    ]
    A = mosflm_matrix[:9]
    unit_cell = uctbx.unit_cell(mosflm_matrix[21:27])
    cm = CrystalFactory.from_mosflm_matrix(A, unit_cell=unit_cell)
    assert approx_equal(
        cm.get_unit_cell().parameters(), unit_cell.parameters(), eps=1.0e-2
    )


def test_crystal_model():
    real_space_a = matrix.col((10, 0, 0))
    real_space_b = matrix.col((0, 11, 0))
    real_space_c = matrix.col((0, 0, 12))
    model = Crystal(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    # This doesn't work as python class uctbx.unit_cell(uctbx_ext.unit_cell)
    # so C++ and python classes are different types
    # assert isinstance(model.get_unit_cell(), uctbx.unit_cell)
    assert model.get_unit_cell().parameters() == (10.0, 11.0, 12.0, 90.0, 90.0, 90.0)
    assert approx_equal(model.get_A(), (1 / 10, 0, 0, 0, 1 / 11, 0, 0, 0, 1 / 12))
    assert approx_equal(
        matrix.sqr(model.get_A()).inverse(), (10, 0, 0, 0, 11, 0, 0, 0, 12)
    )
    assert approx_equal(model.get_B(), model.get_A())
    assert approx_equal(model.get_U(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(
        model.get_real_space_vectors(), (real_space_a, real_space_b, real_space_c)
    )
    assert (
        model.get_crystal_symmetry().unit_cell().parameters()
        == model.get_unit_cell().parameters()
    )
    assert model.get_crystal_symmetry().space_group() == model.get_space_group()

    model2 = Crystal(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    assert model == model2

    model2a = Crystal(model.get_A(), model.get_space_group())
    assert model == model2a

    model2b = Crystal(
        matrix.sqr(model.get_A()).inverse().elems,
        model.get_space_group().type().lookup_symbol(),
        reciprocal=False,
    )
    assert model == model2b

    # rotate 45 degrees about x-axis
    R1 = matrix.sqr(
        (
            1,
            0,
            0,
            0,
            math.cos(math.pi / 4),
            -math.sin(math.pi / 4),
            0,
            math.sin(math.pi / 4),
            math.cos(math.pi / 4),
        )
    )
    # rotate 30 degrees about y-axis
    R2 = matrix.sqr(
        (
            math.cos(math.pi / 6),
            0,
            math.sin(math.pi / 6),
            0,
            1,
            0,
            -math.sin(math.pi / 6),
            0,
            math.cos(math.pi / 6),
        )
    )
    # rotate 60 degrees about z-axis
    R3 = matrix.sqr(
        (
            math.cos(math.pi / 3),
            -math.sin(math.pi / 3),
            0,
            math.sin(math.pi / 3),
            math.cos(math.pi / 3),
            0,
            0,
            0,
            1,
        )
    )
    R = R1 * R2 * R3
    model.set_U(R)
    # B is unchanged
    assert approx_equal(model.get_B(), (1 / 10, 0, 0, 0, 1 / 11, 0, 0, 0, 1 / 12))
    assert approx_equal(model.get_U(), R)
    assert approx_equal(
        model.get_A(), matrix.sqr(model.get_U()) * matrix.sqr(model.get_B())
    )
    a_, b_, c_ = model.get_real_space_vectors()
    assert approx_equal(a_, R * real_space_a)
    assert approx_equal(b_, R * real_space_b)
    assert approx_equal(c_, R * real_space_c)
    assert (
        str(model).replace("-0.000000", " 0.000000")
        == """\
Crystal:
    Unit cell: 10.000, 11.000, 12.000, 90.000, 90.000, 90.000
    Space group: P 1
    U matrix:  {{ 0.433013, -0.750000,  0.500000},
                { 0.789149,  0.047367, -0.612372},
                { 0.435596,  0.659740,  0.612372}}
    B matrix:  {{ 0.100000,  0.000000,  0.000000},
                { 0.000000,  0.090909,  0.000000},
                { 0.000000,  0.000000,  0.083333}}
    A = UB:    {{ 0.043301, -0.068182,  0.041667},
                { 0.078915,  0.004306, -0.051031},
                { 0.043560,  0.059976,  0.051031}}"""
    )
    model.set_B((1 / 12, 0, 0, 0, 1 / 12, 0, 0, 0, 1 / 12))
    assert approx_equal(model.get_unit_cell().parameters(), (12, 12, 12, 90, 90, 90))

    U = matrix.sqr(
        (0.3455, -0.2589, -0.9020, 0.8914, 0.3909, 0.2293, 0.2933, -0.8833, 0.3658)
    )
    B = matrix.sqr((1 / 13, 0, 0, 0, 1 / 13, 0, 0, 0, 1 / 13))
    model.set_A(U * B)
    assert approx_equal(model.get_A(), U * B)
    assert approx_equal(model.get_U(), U, 1e-4)
    assert approx_equal(model.get_B(), B, 1e-5)

    model3 = Crystal(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group=sgtbx.space_group_info("P 222").group(),
    )
    assert model3.get_space_group().type().hall_symbol() == " P 2 2"
    assert model != model3
    #
    sgi_ref = sgtbx.space_group_info(number=230)
    model_ref = Crystal(
        real_space_a=(44, 0, 0),
        real_space_b=(0, 44, 0),
        real_space_c=(0, 0, 44),
        space_group=sgi_ref.group(),
    )
    assert approx_equal(model_ref.get_U(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(model_ref.get_B(), (1 / 44, 0, 0, 0, 1 / 44, 0, 0, 0, 1 / 44))
    assert approx_equal(model_ref.get_A(), model_ref.get_B())
    assert approx_equal(
        model_ref.get_unit_cell().parameters(), (44, 44, 44, 90, 90, 90)
    )
    a_ref, b_ref, c_ref = map(matrix.col, model_ref.get_real_space_vectors())
    cb_op_to_primitive = sgi_ref.change_of_basis_op_to_primitive_setting()
    model_primitive = model_ref.change_basis(cb_op_to_primitive)
    cb_op_to_reference = (
        model_primitive.get_space_group()
        .info()
        .change_of_basis_op_to_reference_setting()
    )
    a_prim, b_prim, c_prim = map(matrix.col, model_primitive.get_real_space_vectors())
    assert (
        cb_op_to_primitive.as_abc()
        == "-1/2*a+1/2*b+1/2*c,1/2*a-1/2*b+1/2*c,1/2*a+1/2*b-1/2*c"
    )
    assert approx_equal(a_prim, -1 / 2 * a_ref + 1 / 2 * b_ref + 1 / 2 * c_ref)
    assert approx_equal(b_prim, 1 / 2 * a_ref - 1 / 2 * b_ref + 1 / 2 * c_ref)
    assert approx_equal(c_prim, 1 / 2 * a_ref + 1 / 2 * b_ref - 1 / 2 * c_ref)
    assert cb_op_to_reference.as_abc() == "b+c,a+c,a+b"
    assert approx_equal(a_ref, b_prim + c_prim)
    assert approx_equal(b_ref, a_prim + c_prim)
    assert approx_equal(c_ref, a_prim + b_prim)
    assert approx_equal(
        model_primitive.get_U(),
        [
            -0.5773502691896258,
            0.40824829046386285,
            0.7071067811865476,
            0.5773502691896257,
            -0.4082482904638631,
            0.7071067811865476,
            0.5773502691896257,
            0.8164965809277259,
            0.0,
        ],
    )
    assert approx_equal(
        model_primitive.get_B(),
        [
            0.0262431940540739,
            0.0,
            0.0,
            0.00927837023781507,
            0.02783511071344521,
            0.0,
            0.01607060866333063,
            0.01607060866333063,
            0.03214121732666125,
        ],
    )
    assert approx_equal(
        model_primitive.get_A(),
        (0, 1 / 44, 1 / 44, 1 / 44, 0, 1 / 44, 1 / 44, 1 / 44, 0),
    )
    assert approx_equal(
        model_primitive.get_unit_cell().parameters(),
        [
            38.1051177665153,
            38.1051177665153,
            38.1051177665153,
            109.47122063449069,
            109.47122063449069,
            109.47122063449069,
        ],
    )
    assert model_ref != model_primitive
    model_ref_recycled = model_primitive.change_basis(cb_op_to_reference)
    assert approx_equal(model_ref.get_U(), model_ref_recycled.get_U())
    assert approx_equal(model_ref.get_B(), model_ref_recycled.get_B())
    assert approx_equal(model_ref.get_A(), model_ref_recycled.get_A())
    assert approx_equal(
        model_ref.get_unit_cell().parameters(),
        model_ref_recycled.get_unit_cell().parameters(),
    )
    assert model_ref == model_ref_recycled

    uc = uctbx.unit_cell((58.2567, 58.1264, 39.7093, 46.9077, 46.8612, 62.1055))
    sg = sgtbx.space_group_info(symbol="P1").group()
    cs = crystal.symmetry(unit_cell=uc, space_group=sg)
    cb_op_to_minimum = cs.change_of_basis_op_to_minimum_cell()
    # the reciprocal matrix
    B = matrix.sqr(uc.fractionalization_matrix()).transpose()
    U = random_rotation()
    direct_matrix = (U * B).inverse()
    model = Crystal(
        direct_matrix[:3], direct_matrix[3:6], direct_matrix[6:9], space_group=sg
    )
    assert uc.is_similar_to(model.get_unit_cell())
    uc_minimum = uc.change_basis(cb_op_to_minimum)
    model_minimum = model.change_basis(cb_op_to_minimum)
    assert uc_minimum.is_similar_to(model_minimum.get_unit_cell())
    assert model_minimum != model
    model_minimum.update(model)
    assert model_minimum == model  # lgtm

    A_static = matrix.sqr(model.get_A())
    A_as_scan_points = [A_static]
    num_scan_points = 11
    for i in range(num_scan_points - 1):
        A_as_scan_points.append(
            A_as_scan_points[-1] * matrix.sqr(euler_angles.xyz_matrix(0.1, 0.2, 0.3))
        )
    model.set_A_at_scan_points(A_as_scan_points)
    model_minimum = model.change_basis(cb_op_to_minimum)
    assert model.num_scan_points == model_minimum.num_scan_points == num_scan_points
    M = matrix.sqr(cb_op_to_minimum.c_inv().r().transpose().as_double())
    M_inv = M.inverse()
    for i in range(num_scan_points):
        A_orig = matrix.sqr(model.get_A_at_scan_point(i))
        A_min = matrix.sqr(model_minimum.get_A_at_scan_point(i))
        assert approx_equal(A_min, A_orig * M_inv)
    assert model.get_unit_cell().parameters() == pytest.approx(
        (58.2567, 58.1264, 39.7093, 46.9077, 46.8612, 62.1055)
    )
    uc = uctbx.unit_cell((10, 11, 12, 91, 92, 93))
    model.set_unit_cell(uc)
    assert model.get_unit_cell().parameters() == pytest.approx(uc.parameters())


def test_MosaicCrystalKabsch2010():
    mosaic_model = MosaicCrystalKabsch2010(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    mosaic_model2 = MosaicCrystalKabsch2010(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    assert (
        str(mosaic_model).replace("-0.000000", " 0.000000")
        == """\
Crystal:
    Unit cell: 10.000, 11.000, 12.000, 90.000, 90.000, 90.000
    Space group: P 1
    U matrix:  {{ 1.000000,  0.000000,  0.000000},
                { 0.000000,  1.000000,  0.000000},
                { 0.000000,  0.000000,  1.000000}}
    B matrix:  {{ 0.100000,  0.000000,  0.000000},
                { 0.000000,  0.090909,  0.000000},
                { 0.000000,  0.000000,  0.083333}}
    A = UB:    {{ 0.100000,  0.000000,  0.000000},
                { 0.000000,  0.090909,  0.000000},
                { 0.000000,  0.000000,  0.083333}}
    Mosaicity:  0.000000"""
    )
    assert approx_equal(mosaic_model.get_mosaicity(), 0)
    assert mosaic_model == mosaic_model2
    mosaic_model2.set_mosaicity(0.01)
    assert mosaic_model != mosaic_model2  # lgtm
    # FIXME Crystal == MosaicCrystal gives unexpected result, depending on
    # parameter order
    # model4 = Crystal(real_space_a=(10,0,0),
    #                  real_space_b=(0,11,0),
    #                  real_space_c=(0,0,12),
    #                  space_group_symbol="P 1")
    # print "model4 == mosaic_model", model4 == mosaic_model # True
    # print "mosaic_model == model4", mosaic_model == model4 # False
    # print "mosaic_model2 == model4", mosaic_model2 == model4 # False
    # print "model4 == mosaic_model2", model4 == mosaic_model2 # True


def test_MosaicCrystalSauter2014():
    mosaic_model = MosaicCrystalSauter2014(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    mosaic_model2 = MosaicCrystalSauter2014(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    assert (
        str(mosaic_model).replace("-0.000000", " 0.000000")
        == """\
Crystal:
    Unit cell: 10.000, 11.000, 12.000, 90.000, 90.000, 90.000
    Space group: P 1
    U matrix:  {{ 1.000000,  0.000000,  0.000000},
                { 0.000000,  1.000000,  0.000000},
                { 0.000000,  0.000000,  1.000000}}
    B matrix:  {{ 0.100000,  0.000000,  0.000000},
                { 0.000000,  0.090909,  0.000000},
                { 0.000000,  0.000000,  0.083333}}
    A = UB:    {{ 0.100000,  0.000000,  0.000000},
                { 0.000000,  0.090909,  0.000000},
                { 0.000000,  0.000000,  0.083333}}
    Half mosaic angle (degrees):  0.000000
    Domain size (Angstroms):  0.000000"""
    )
    assert approx_equal(mosaic_model.get_half_mosaicity_deg(), 0)
    assert approx_equal(mosaic_model.get_domain_size_ang(), 0)
    assert mosaic_model == mosaic_model2
    mosaic_model2.set_half_mosaicity_deg(0.01)
    assert mosaic_model != mosaic_model2  # lgtm
    mosaic_model2.set_half_mosaicity_deg(0)
    assert mosaic_model == mosaic_model2  # lgtm
    mosaic_model2.set_domain_size_ang(1000)
    assert mosaic_model != mosaic_model2  # lgtm


def test_similarity():
    model_1 = MosaicCrystalKabsch2010(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    model_1.set_mosaicity(0.5)
    model_2 = MosaicCrystalKabsch2010(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    model_2.set_mosaicity(0.5)
    assert model_1.is_similar_to(model_2)
    model_1.set_mosaicity(-1)
    model_2.set_mosaicity(-0.5)
    assert model_1.is_similar_to(model_2)  # test ignores negative mosaicity
    model_1.set_mosaicity(0.5)
    model_2.set_mosaicity(0.63)  # outside tolerance
    assert not model_1.is_similar_to(model_2)
    model_2.set_mosaicity(0.62)  # just inside tolerance

    # orientation tests
    R = matrix.sqr(model_2.get_U())
    dr1 = matrix.col((1, 0, 0)).axis_and_angle_as_r3_rotation_matrix(0.0101, deg=True)
    dr2 = matrix.col((1, 0, 0)).axis_and_angle_as_r3_rotation_matrix(0.0099, deg=True)
    model_2.set_U(dr1 * R)
    assert not model_1.is_similar_to(model_2)  # outside tolerance
    model_2.set_U(dr2 * R)
    assert model_1.is_similar_to(model_2)  # inside tolerance

    # unit_cell.is_similar_to is tested elsewhere


def test_change_basis_mosaic_crystal():
    mosaic_model = MosaicCrystalSauter2014(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )
    mosaic_model.set_half_mosaicity_deg(0.01)
    mosaic_model.set_domain_size_ang(1000)
    assert mosaic_model.get_half_mosaicity_deg() == 0.01
    assert mosaic_model.get_domain_size_ang() == 1000

    # Currently change_basis downgrades the MosaicCrystalSauter2014 to
    # a Crystal object so this will fail
    mosaic_model2 = mosaic_model.change_basis(change_of_basis_op())
    assert mosaic_model2.get_half_mosaicity_deg() == 0.01
    assert mosaic_model2.get_domain_size_ang() == 1000
    assert mosaic_model.is_similar_to(mosaic_model2)


def test_check_old_vs_new():
    model_1 = Crystal(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )

    model_2 = crystal_model_old(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )

    cov_B = matrix.sqr([1] * (9 * 9))

    model_1.set_B_covariance(cov_B)
    model_2.set_B_covariance(cov_B)

    A_list = [model_1.get_A() for i in range(20)]

    model_1.set_A_at_scan_points(A_list)
    model_2.set_A_at_scan_points(A_list)

    A1 = model_1.get_A()
    A2 = model_2.get_A()
    U1 = model_1.get_U()
    U2 = model_2.get_U()
    B1 = model_1.get_B()
    B2 = model_2.get_B()
    UC1 = model_1.get_unit_cell()
    UC2 = model_2.get_unit_cell()
    RSV1 = model_1.get_real_space_vectors()
    RSV2 = model_2.get_real_space_vectors()
    SG1 = model_1.get_space_group()
    SG2 = model_2.get_space_group()

    assert model_1.num_scan_points == model_2.num_scan_points

    A_list_1 = [
        model_1.get_A_at_scan_point(i) for i in range(model_1.get_num_scan_points())
    ]
    A_list_2 = [
        model_2.get_A_at_scan_point(i) for i in range(model_1.get_num_scan_points())
    ]
    B_list_1 = [
        model_1.get_B_at_scan_point(i) for i in range(model_1.get_num_scan_points())
    ]
    B_list_2 = [
        model_2.get_B_at_scan_point(i) for i in range(model_1.get_num_scan_points())
    ]
    U_list_1 = [
        model_1.get_U_at_scan_point(i) for i in range(model_1.get_num_scan_points())
    ]
    U_list_2 = [
        model_2.get_U_at_scan_point(i) for i in range(model_1.get_num_scan_points())
    ]

    assert approx_equal(A1, A2)
    assert approx_equal(B1, B2)
    assert approx_equal(U1, U2)
    assert approx_equal(UC1.parameters(), UC2.parameters())
    assert approx_equal(RSV1[0], RSV2[0])
    assert approx_equal(RSV1[1], RSV2[1])
    assert approx_equal(RSV1[2], RSV2[2])
    assert str(SG1.info()) == str(SG2.info())

    for i in range(model_1.get_num_scan_points()):
        assert approx_equal(A_list_1[i], A_list_2[i])
        assert approx_equal(B_list_1[i], B_list_2[i])
        assert approx_equal(U_list_1[i], U_list_2[i])

    cell_sd_1 = model_1.get_cell_parameter_sd()
    cell_sd_2 = model_2.get_cell_parameter_sd()
    cell_volume_sd_1 = model_1.get_cell_volume_sd()
    cell_volume_sd_2 = model_2.get_cell_volume_sd()
    covB1 = model_1.get_B_covariance()
    covB2 = model_1.get_B_covariance()

    assert approx_equal(covB1, covB2)
    assert approx_equal(cell_volume_sd_1, cell_volume_sd_2)
    assert approx_equal(cell_sd_1, cell_sd_2)


@pytest.mark.parametrize(
    "crystal_class", [Crystal, MosaicCrystalKabsch2010, MosaicCrystalSauter2014]
)
def test_set_scan_varying_B_covariance(crystal_class):
    xl = crystal_class(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )

    cov_B = flex.double(
        (
            [8e-14, -1e-29, 3e-30, 3e-14, 8e-14, 3e-30, 2e-15, -7e-15, 2e-14],
            [-1e-29, 2e-45, -4e-46, -4e-30, -1e-29, -4e-46, 1e-30, 4e-30, -3e-30],
            [3e-30, -4e-46, 1e-46, 9e-31, 3e-30, 1e-46, 2e-31, -2e-31, 9e-31],
            [3e-14, -4e-30, 9e-31, 4e-14, 3e-14, 9e-31, -2e-15, -1e-15, 7e-15],
            [8e-14, -1e-29, 3e-30, 3e-14, 1e-13, 3e-30, 6e-15, -1e-15, 3e-14],
            [3e-30, -4e-46, 1e-46, 9e-31, 3e-30, 1e-46, 2e-31, -2e-31, 9e-31],
            [2e-15, 1e-30, 2e-31, -2e-15, 6e-15, 2e-31, 2e-14, 1e-14, 2e-15],
            [-7e-15, 4e-30, -2e-31, -1e-15, -1e-15, -2e-31, 1e-14, 3e-14, -2e-15],
            [2e-14, -3e-30, 9e-31, 7e-15, 3e-14, 9e-31, 2e-15, -2e-15, 8e-15],
        )
    )
    xl.set_B_covariance(cov_B)

    cov_B_array = flex.double(flex.grid(20, 9, 9))
    cov_B_2d = cov_B.as_scitbx_matrix()
    cov_B.reshape(flex.grid(1, 9, 9))
    for i in range(20):
        cov_B_array[i : (i + 1), :, :] = cov_B

    # This should fail: set_A_at_scan_points should be called first
    with pytest.raises(RuntimeError):
        xl.set_B_covariance_at_scan_points(cov_B_array)
        pytest.fail("Setting B covariance before A at scan points is not allowed")

    A_list = [xl.get_A() for i in range(20)]
    xl.set_A_at_scan_points(A_list)

    # Now the setter should work
    xl.set_B_covariance_at_scan_points(cov_B_array)

    # Check getters for B covariance and cell sd
    cell_sd = xl.get_cell_parameter_sd()
    for i in range(20):
        cov_B_at_scan_point = xl.get_B_covariance_at_scan_point(i)
        assert cov_B_at_scan_point == cov_B_2d
        cell_sd_at_scan_point = xl.get_cell_parameter_sd_at_scan_point(i)
        assert cell_sd_at_scan_point == pytest.approx(cell_sd)


@pytest.mark.parametrize(
    "crystal_class", [Crystal, MosaicCrystalKabsch2010, MosaicCrystalSauter2014]
)
def test_recalculated_cell(crystal_class):
    xl = crystal_class(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group_symbol="P 1",
    )

    assert xl.get_recalculated_unit_cell() is None
    assert xl.get_recalculated_cell_volume_sd() == 0

    uc1 = xl.get_unit_cell()
    uc2 = uctbx.unit_cell((10, 11, 10, 90, 90, 90))
    xl.set_recalculated_unit_cell(uc2)
    uc3 = xl.get_recalculated_unit_cell()
    assert uc3 is not None
    assert not uc1.is_similar_to(uc2)
    assert uc2.is_similar_to(uc3)

    xl.set_recalculated_cell_parameter_sd((0.1,) * 6)
    xl.set_recalculated_cell_volume_sd(0.001)
    assert xl.get_recalculated_cell_parameter_sd() == (0.1, 0.1, 0.1, 0.1, 0.1, 0.1)
    assert xl.get_recalculated_cell_volume_sd() == 0.001
    assert (
        "    Recalculated unit cell: 10.00(10)"
        ", 11.00(10), 10.00(10), 90.00(10), 90.00(10), 90.00(10)"
        in str(xl).splitlines()
    )

    cb_op = sgtbx.change_of_basis_op("b,c,a")
    assert xl.change_basis(
        cb_op
    ).get_recalculated_unit_cell().parameters() == pytest.approx(
        (11.0, 10.0, 10.0, 90.0, 90.0, 90.0)
    )

    # Verify that set_recalculated_unit_cell() resets the cell parameter sd's
    xl.set_recalculated_unit_cell(uctbx.unit_cell((11, 12, 13, 90, 90, 90)))
    assert xl.get_recalculated_cell_parameter_sd() == ()
    assert xl.get_recalculated_cell_volume_sd() == 0


def test_from_phil():
    params = crystal_phil_scope.extract()
    assert CrystalFactory.from_phil(params) is None

    params = crystal_phil_scope.fetch(
        parse(
            """
    crystal {
      unit_cell = 10, 20, 30, 90, 90, 90
      space_group = P222
    }
  """
        )
    ).extract()

    # Create the crystal
    crystal = CrystalFactory.from_phil(params)
    assert crystal.get_unit_cell().parameters() == (10, 20, 30, 90, 90, 90)
    assert crystal.get_space_group().type().lookup_symbol() == "P 2 2 2"

    params = crystal_phil_scope.fetch(
        parse(
            """
    crystal {
      A_matrix =  0.0541,  0.0832, -0.0060, 0.0049, -0.0175, -0.0492, -0.0840,  0.0526, -0.0068
      space_group = P4
    }
  """
        )
    ).extract()

    # Create the crystal
    crystal = CrystalFactory.from_phil(params)
    assert crystal.get_A() == pytest.approx(
        (0.0541, 0.0832, -0.0060, 0.0049, -0.0175, -0.0492, -0.0840, 0.0526, -0.0068)
    )
    assert crystal.get_space_group().type().lookup_symbol() == "P 4"



from __future__ import annotations

import pickle

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format


def test_pickle_incorrect_format():
    ex = IncorrectFormatError(Format, "Some message")
    pickle.loads(pickle.dumps(ex))



from __future__ import annotations

import math
from pathlib import Path

import pytest

from libtbx import easy_pickle
from libtbx.phil import parse
from libtbx.test_utils import Exception_expected, approx_equal
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.model.goniometer import (
    Goniometer,
    GoniometerFactory,
    KappaGoniometer,
    MultiAxisGoniometer,
    goniometer_phil_scope,
)


def _compare_tuples(a, b, tolerance=1.0e-6):
    assert len(a) == len(b)

    for va, vb in zip(a, b):
        assert math.fabs(va - vb) <= tolerance


def test_goniometer():
    """A test class for the goniometer class."""

    axis = (1, 0, 0)
    fixed = (1, 0, 0, 0, 1, 0, 0, 0, 1)

    xg = Goniometer(axis, fixed)

    assert len(xg.get_rotation_axis()) == 3
    assert len(xg.get_fixed_rotation()) == 9

    _compare_tuples(xg.get_rotation_axis(), axis)
    _compare_tuples(xg.get_fixed_rotation(), fixed)

    single = GoniometerFactory.single_axis()

    assert len(single.get_rotation_axis()) == 3
    assert len(single.get_fixed_rotation()) == 9

    _compare_tuples(single.get_rotation_axis(), axis)
    _compare_tuples(single.get_fixed_rotation(), fixed)

    kappa = GoniometerFactory.kappa(50.0, 0.0, 0.0, 0.0, "-y", "omega")

    assert len(kappa.get_rotation_axis()) == 3
    assert len(kappa.get_fixed_rotation()) == 9
    _compare_tuples(kappa.get_rotation_axis(), axis)
    _compare_tuples(kappa.get_fixed_rotation(), fixed)

    kappa = GoniometerFactory.kappa(50.0, 0.0, 0.0, 0.0, "-y", "omega")

    assert len(kappa.get_rotation_axis()) == 3
    assert len(kappa.get_fixed_rotation()) == 9

    _compare_tuples(kappa.get_rotation_axis(), axis)
    _compare_tuples(kappa.get_fixed_rotation(), fixed)

    kappa = GoniometerFactory.kappa(50.0, 0.0, 0.0, 0.0, "-y", "phi")

    assert len(kappa.get_rotation_axis()) == 3
    assert len(kappa.get_fixed_rotation()) == 9

    _compare_tuples(kappa.get_rotation_axis(), axis)
    _compare_tuples(kappa.get_fixed_rotation(), fixed)

    kappa = GoniometerFactory.kappa(50.0, 0.0, 30.0, 0.0, "-y", "omega")

    assert len(kappa.get_rotation_axis()) == 3
    assert len(kappa.get_fixed_rotation()) == 9

    _compare_tuples(kappa.get_rotation_axis(), axis)
    with pytest.raises(AssertionError):
        _compare_tuples(kappa.get_fixed_rotation(), fixed)

    image = Path(__file__).parent / "phi_scan_001.cbf"
    assert GoniometerFactory.imgCIF(str(image))

    kappa = GoniometerFactory.kappa(50.0, -10.0, 30.0, 0.0, "-y", "phi")

    s = easy_pickle.dumps(kappa)
    kappa2 = easy_pickle.loads(s)
    assert kappa == kappa2

    image = Path(__file__).parent / "omega_scan.cbf"
    assert GoniometerFactory.imgCIF(str(image))

    kappa = GoniometerFactory.kappa(50.0, -10.0, 30.0, 20.0, "-y", "omega")

    s = easy_pickle.dumps(kappa)
    kappa2 = easy_pickle.loads(s)
    assert kappa == kappa2


def test_multi_axis_goniometer():
    alpha = 50
    omega = -10
    kappa = 30
    phi = 20
    direction = "-y"

    kappa_omega_scan = KappaGoniometer(alpha, omega, kappa, phi, direction, "omega")
    axes = (
        kappa_omega_scan.get_phi_axis(),
        kappa_omega_scan.get_kappa_axis(),
        kappa_omega_scan.get_omega_axis(),
    )
    angles = (
        kappa_omega_scan.get_phi_angle(),
        kappa_omega_scan.get_kappa_angle(),
        kappa_omega_scan.get_omega_angle(),
    )

    # First test a kappa goniometer with omega as the scan axis
    axes = flex.vec3_double(axes)
    angles = flex.double(angles)
    names = flex.std_string(("phi", "kappa", "omega"))
    scan_axis = 2

    multi_axis_omega_scan = GoniometerFactory.multi_axis(axes, angles, names, scan_axis)
    assert approx_equal(
        multi_axis_omega_scan.get_fixed_rotation(),
        kappa_omega_scan.get_fixed_rotation(),
    )
    assert approx_equal(
        multi_axis_omega_scan.get_rotation_axis(), kappa_omega_scan.get_rotation_axis()
    )

    recycle_omega = MultiAxisGoniometer.from_dict(multi_axis_omega_scan.to_dict())
    assert approx_equal(recycle_omega.get_axes(), multi_axis_omega_scan.get_axes())
    assert approx_equal(recycle_omega.get_angles(), multi_axis_omega_scan.get_angles())
    assert recycle_omega.get_scan_axis() == multi_axis_omega_scan.get_scan_axis()

    # Now test a kappa goniometer with phi as the scan axis
    kappa_phi_scan = KappaGoniometer(alpha, omega, kappa, phi, direction, "phi")

    scan_axis = 0
    multi_axis_phi_scan = GoniometerFactory.multi_axis(axes, angles, names, scan_axis)
    assert approx_equal(
        multi_axis_phi_scan.get_fixed_rotation(), kappa_phi_scan.get_fixed_rotation()
    )
    assert approx_equal(
        matrix.sqr(multi_axis_phi_scan.get_setting_rotation())
        * multi_axis_phi_scan.get_rotation_axis_datum(),
        kappa_phi_scan.get_rotation_axis(),
    )
    assert approx_equal(
        multi_axis_phi_scan.get_rotation_axis(), kappa_phi_scan.get_rotation_axis()
    )

    recycle_phi = MultiAxisGoniometer.from_dict(multi_axis_phi_scan.to_dict())
    assert approx_equal(recycle_phi.get_axes(), multi_axis_phi_scan.get_axes())
    assert approx_equal(recycle_phi.get_angles(), multi_axis_phi_scan.get_angles())
    assert recycle_phi.get_scan_axis() == multi_axis_phi_scan.get_scan_axis()

    s = easy_pickle.dumps(multi_axis_phi_scan)
    recycle = easy_pickle.loads(s)
    assert recycle == multi_axis_phi_scan

    assert approx_equal(recycle.get_axes(), multi_axis_phi_scan.get_axes())
    assert approx_equal(recycle.get_angles(), multi_axis_phi_scan.get_angles())
    assert recycle.get_scan_axis() == multi_axis_phi_scan.get_scan_axis()
    recycle.set_angles((0, 90, 180))
    assert approx_equal(recycle.get_angles(), (0, 90, 180))
    new_axes = flex.vec3_double(
        (
            (0.99996, -0.00647, -0.00659),
            (0.91314, 0.27949, -0.29674),
            (1.00000, -0.00013, -0.00064),
        )
    )
    recycle.set_axes(new_axes)
    assert approx_equal(recycle.get_axes(), new_axes)

    # Check exception is raised if scan axis is out range
    try:
        GoniometerFactory.multi_axis(axes, angles, names, 3)
    except RuntimeError:
        pass
    else:
        raise Exception_expected

    # Single axis is just a special case of a multi axis goniometer
    single_axis = GoniometerFactory.multi_axis(
        flex.vec3_double(((1, 0, 0),)), flex.double((0,)), flex.std_string(("PHI",)), 0
    )
    assert single_axis.get_fixed_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert single_axis.get_setting_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert single_axis.get_rotation_axis() == (1, 0, 0)


def test_single_axis_goniometer_from_phil():
    params = goniometer_phil_scope.fetch(
        parse(
            """
    goniometer {
      axis = (1, 0, 0)
    }
  """
        )
    ).extract()

    g1 = GoniometerFactory.from_phil(params)

    assert g1.get_rotation_axis() == (1, 0, 0)


def test_single_axis_goniometer_using_axes_from_phil():
    params = goniometer_phil_scope.fetch(
        parse(
            """
    goniometer {
      axes = (1, 0, 0)
    }
  """
        )
    ).extract()

    g1 = GoniometerFactory.from_phil(params)

    assert g1.get_rotation_axis() == (1, 0, 0)


def test_single_axis_goniometer_using_axis_and_axes_from_phil_raises_error():
    """Supplying both the 'axis' and 'axes' parameters is ambiguous, so it should
    be mutually exclusive. This test ensures that is the case."""

    params = goniometer_phil_scope.fetch(
        parse(
            """
    goniometer {
      axis = (1, 0, 0)
      axes = (1, 0, 0)
    }
    """
        )
    ).extract()

    with pytest.raises(ValueError):
        GoniometerFactory.from_phil(params)


def test_single_axis_goniometer_with_fixed_rotation_and_reference_from_phil():
    params = goniometer_phil_scope.fetch(
        parse(
            """
    goniometer {
      axes = (1, 0, 0)
    }
  """
        )
    ).extract()

    g1 = GoniometerFactory.from_phil(params)

    params = goniometer_phil_scope.fetch(
        parse(
            """
    goniometer {
      axis = (0, 1, 0)
      fixed_rotation = (0, 1, 0, 1, 0, 0, 0, 0, 1)
    }
  """
        )
    ).extract()

    g2 = GoniometerFactory.from_phil(params, reference=g1)

    assert g2.get_rotation_axis() == (0, 1, 0)
    assert g2.get_fixed_rotation() == (0, 1, 0, 1, 0, 0, 0, 0, 1)


def test_multi_axis_goniometer_from_phil():
    params = goniometer_phil_scope.fetch(
        parse(
            """
    goniometer {
      axes = (1, 0, 0, 0, 1, 0, 0, 0, 1)
      scan_axis = 2
    }
  """
        )
    ).extract()

    g1 = GoniometerFactory.from_phil(params)
    assert tuple(g1.get_axes()) == ((1, 0, 0), (0, 1, 0), (0, 0, 1))
    assert g1.get_scan_axis() == 2

    # Test using a reference
    params = goniometer_phil_scope.fetch(
        parse(
            """
    goniometer {
      axes = (0, 1, 0, 1, 0, 0, 0, 0, 1)
    }
  """
        )
    ).extract()

    g2 = GoniometerFactory.from_phil(params, reference=g1)

    assert tuple(g2.get_axes()) == ((0, 1, 0), (1, 0, 0), (0, 0, 1))


def test_scan_varying():
    axis = (1, 0, 0)
    g = Goniometer(axis)

    assert g.get_num_scan_points() == 0
    assert g.get_setting_rotation_at_scan_points().size() == 0
    with pytest.raises(RuntimeError):
        g.get_setting_rotation_at_scan_point(0)

    # set varying beam
    num_scan_points = 11
    S_static = matrix.sqr(g.get_setting_rotation())
    S_as_scan_points = [S_static]
    axis = matrix.col.random(3, -1.0, 1.0).normalize()
    R = axis.axis_and_angle_as_r3_rotation_matrix(angle=0.01, deg=True)
    for i in range(num_scan_points - 1):
        S_as_scan_points.append(R * S_as_scan_points[-1])
    g.set_setting_rotation_at_scan_points(S_as_scan_points)
    assert g.get_num_scan_points() == 11
    assert g.get_setting_rotation_at_scan_points().size() == 11

    for t in range(num_scan_points):
        S_t = matrix.sqr(g.get_setting_rotation_at_scan_point(t))
        assert S_t == S_as_scan_points[t]

    # also test setting as tuple
    g.set_setting_rotation_at_scan_points(tuple(S_as_scan_points))
    assert g.get_num_scan_points() == 11
    assert g.get_setting_rotation_at_scan_points().size() == 11

    # test resetting
    g.reset_scan_points()
    assert g.get_num_scan_points() == 0
    assert g.get_setting_rotation_at_scan_points().size() == 0


def test_comparison():
    # Setting rotation for small random offset
    offset_ax = matrix.col.random(3, -1.0, 1.0).normalize()
    S = offset_ax.axis_and_angle_as_r3_rotation_matrix(angle=0.01, deg=True)

    # Equal goniometers with scan-points set
    g1 = Goniometer((1, 0, 0))
    g1.set_setting_rotation(S)
    g1.set_setting_rotation_at_scan_points([S] * 5)
    g2 = Goniometer((1, 0, 0))
    g2.set_setting_rotation(S)
    g2.set_setting_rotation_at_scan_points([S] * 5)

    assert g1 == g2
    assert g1.is_similar_to(g2)

    # Different setting matrix
    g3 = Goniometer((1, 0, 0))
    invS = S.inverse()
    g3.set_setting_rotation(invS)
    g3.set_setting_rotation_at_scan_points([invS] * 5)

    assert g1 != g3
    assert not g1.is_similar_to(g3)



from __future__ import annotations

try:
    from .dxtbx_flumpy import *  # noqa: F403
except ModuleNotFoundError:
    from dxtbx_flumpy import *  # type: ignore # noqa: F403

__all__ = [  # noqa: F405
    "to_numpy",
    "from_numpy",
    "vec_from_numpy",
    "mat3_from_numpy",
    "Scuffer",
]



from __future__ import annotations

import bz2
import collections
import errno
import os
import pickle
import shutil
from unittest import mock

import dateutil.parser
import pytest

from cctbx import sgtbx
from scitbx.array_family import flex

import dxtbx
import dxtbx.model.experiment_list
from dxtbx.format.Format import Format
from dxtbx.imageset import ImageSetFactory
from dxtbx.model import (
    Beam,
    BeamFactory,
    Crystal,
    Detector,
    Experiment,
    ExperimentList,
    ExperimentType,
    Goniometer,
    History,
    Scan,
    ScanFactory,
)

try:
    from ..dxtbx_model_ext import Probe
except ModuleNotFoundError:
    from dxtbx_model_ext import Probe  # type: ignore
from dxtbx.model.experiment_list import ExperimentListDict, ExperimentListFactory


@pytest.fixture(scope="session")
def centroid_test_data(dials_data):
    return str(dials_data("centroid_test_data"))


@pytest.fixture
def single_sequence_filenames(centroid_test_data):
    filenames = [
        os.path.join(centroid_test_data, f"centroid_000{i}.cbf") for i in range(1, 10)
    ]
    return filenames


@pytest.fixture
def multiple_sequence_filenames(centroid_test_data):
    filenames = [
        os.path.join(centroid_test_data, f"centroid_000{i}.cbf")
        for i in [1, 2, 3, 7, 8, 9]
    ]
    return filenames


@pytest.fixture
def all_image_examples(dials_data):
    filenames = (
        ("ALS_1231-q315r_lyso_1_001.img.bz2"),
        ("ALS_501-als501_q4_1_001.img.bz2"),
        ("ALS_821-q210_lyso_1_101.img.bz2"),
        ("ALS_831-q315r_lyso_001.img.bz2"),
        ("APS_14BMC-q315_1_001.img.bz2"),
        ("APS_17ID-q210_1_001.img.bz2"),
        ("APS_19ID-q315_unbinned_a.0001.img.bz2"),
        ("APS_22ID-mar300.0001"),
        ("APS_23IDD-mar300_1_E1.0001.bz2"),
        ("APS_24IDC-pilatus_1_0001.cbf.bz2"),
        ("APS_24IDC-q315_1_001.img.bz2"),
        ("CLS1_08ID1-mar225_2_E0_0001.img.bz2"),
        ("DESY_ID141-q210_2_001.img.bz2"),
        ("ESRF_BM14-mar165_001.mccd.bz2"),
        ("ESRF_BM14-mar225_1_001.mccd.bz2"),
        ("ESRF_ID231-q315r_7_001.img.bz2"),
        ("RAXIS-HTC-test1_lysozyme_0111060001.osc.bz2"),
        ("SLS_X06SA-mar225_2_001.img.bz2"),
        ("SLS_X06SA-pilatus6m_1_00001.cbf.bz2"),
        ("SRS_101-mar225_001.img.bz2"),
        ("SRS_142-q4_1_001.img.bz2"),
        ("SSRL_bl111-mar325_1_001.mccd.bz2"),
        ("xia2-merge2cbf_averaged_0001.cbf.bz2"),
    )
    return [str(dials_data("image_examples") / f) for f in filenames]


@pytest.fixture
def multiple_block_filenames(single_sequence_filenames, all_image_examples):
    return single_sequence_filenames + all_image_examples


def test_experiment_list_extend():
    """Check that the extend method of ExperimentList works."""
    # Create a minimal ExperimentList instance.
    expts = ExperimentList([Experiment()])
    # Try to extend it.
    expts.extend(expts)


def test_experiment_contains():
    # Create a load of models
    b1 = Beam()
    d1 = Detector()
    g1 = Goniometer()
    s1 = Scan()
    c1 = Crystal((1, 0, 0), (0, 1, 0), (0, 0, 1), space_group_symbol="P1")

    # Create an experiment
    e = Experiment(
        beam=b1, detector=d1, goniometer=g1, scan=s1, crystal=c1, imageset=None
    )

    # Check experiment contains model
    assert b1 in e
    assert d1 in e
    assert g1 in e
    assert s1 in e
    assert c1 in e

    # Create a load of models that look the same but aren't
    b2 = Beam()
    d2 = Detector()
    g2 = Goniometer()
    s2 = Scan()
    c2 = Crystal((1, 0, 0), (0, 1, 0), (0, 0, 1), space_group_symbol="P1")

    # Check experiment doesn't contain model
    assert b2 not in e
    assert d2 not in e
    assert g2 not in e
    assert s2 not in e
    assert c2 not in e


def test_experiment_equality():
    # Create a load of models
    b1 = Beam()
    d1 = Detector()
    g1 = Goniometer()
    s1 = Scan()
    c1 = Crystal((1, 0, 0), (0, 1, 0), (0, 0, 1), space_group_symbol="P1")

    # Create a load of models that look the same but aren't
    b2 = Beam()
    d2 = Detector()
    g2 = Goniometer()
    s2 = Scan()
    c2 = Crystal((1, 0, 0), (0, 1, 0), (0, 0, 1), space_group_symbol="P1")

    # Create a few experiments
    e1 = Experiment(
        beam=b1, detector=d1, goniometer=g1, scan=s1, crystal=c1, imageset=None
    )
    e2 = Experiment(
        beam=b1, detector=d1, goniometer=g1, scan=s1, crystal=c1, imageset=None
    )
    e3 = Experiment(
        beam=b2, detector=d2, goniometer=g2, scan=s2, crystal=c2, imageset=None
    )

    # Check e1 equals e2 and neither equals e3
    assert e1 == e2
    assert e1 != e3
    assert e2 != e3


def test_experiment_consistent(dials_data):
    # Create a sequence
    sequence_filenames = dials_data("centroid_test_data").glob("centroid*.cbf")
    sequence = ImageSetFactory.new(sorted(str(f) for f in sequence_filenames))[0]

    # Create experiment with sequence and good scan
    e = Experiment(imageset=sequence, scan=sequence.get_scan())
    assert e.is_consistent()

    # Create experiment with sequence and defective scan
    scan = sequence.get_scan()
    scan.set_image_range((1, 1))
    e = Experiment(imageset=sequence, scan=scan)
    # assert not e.is_consistent()) # FIXME

    ## Create experiment with imageset and good scan
    # assert e.is_consistent()

    ## Create experiment with imageset and non-still scan
    # assert not e.is_consistent()

    ## Create experiment with imageset and scan with more than 1 image
    # assert not e.is_consistent()

    ## Create experiment with imageset and defective scan
    # assert not e.is_consistent()


def test_experimentlist_contains(experiment_list):
    # Check all the models are found
    for e in experiment_list:
        assert e.beam in experiment_list
        assert e.detector in experiment_list
        assert e.goniometer in experiment_list
        assert e.scan in experiment_list

    # Create some more models
    b = Beam()
    d = Detector()
    g = Goniometer()
    s = Scan()

    # Check that models not in are not found
    assert b not in experiment_list
    assert d not in experiment_list
    assert g not in experiment_list
    assert s not in experiment_list


def test_experimentlist_replace(experiment_list):
    # Get the models
    b = [e.beam for e in experiment_list]

    # Replace some models
    experiment_list.replace(b[0], b[1])
    assert experiment_list[0].beam is b[1]
    assert experiment_list[4].beam is b[1]

    # Replace again
    experiment_list[0].beam = b[0]
    experiment_list[4].beam = b[4]


def test_experimentlist_indices(experiment_list):
    # Get the models
    b = [e.beam for e in experiment_list]
    d = [e.detector for e in experiment_list]
    g = [e.goniometer for e in experiment_list]
    s = [e.scan for e in experiment_list]

    # Check indices of beams
    assert list(experiment_list.indices(b[0])) == [0, 4]
    assert list(experiment_list.indices(b[1])) == [1, 3]
    assert list(experiment_list.indices(b[2])) == [2]
    assert list(experiment_list.indices(b[3])) == [1, 3]
    assert list(experiment_list.indices(b[4])) == [0, 4]

    # Check indices of detectors
    assert list(experiment_list.indices(d[0])) == [0, 4]
    assert list(experiment_list.indices(d[1])) == [1, 3]
    assert list(experiment_list.indices(d[2])) == [2]
    assert list(experiment_list.indices(d[3])) == [1, 3]
    assert list(experiment_list.indices(d[4])) == [0, 4]

    # Check indices of goniometer
    assert list(experiment_list.indices(g[0])) == [0, 4]
    assert list(experiment_list.indices(g[1])) == [1, 3]
    assert list(experiment_list.indices(g[2])) == [2]
    assert list(experiment_list.indices(g[3])) == [1, 3]
    assert list(experiment_list.indices(g[4])) == [0, 4]

    # Check indices of scans
    assert list(experiment_list.indices(s[0])) == [0, 4]
    assert list(experiment_list.indices(s[1])) == [1, 3]
    assert list(experiment_list.indices(s[2])) == [2]
    assert list(experiment_list.indices(s[3])) == [1, 3]
    assert list(experiment_list.indices(s[4])) == [0, 4]

    # Check some models not in the list
    assert len(experiment_list.indices(Beam())) == 0
    assert len(experiment_list.indices(Detector())) == 0
    assert len(experiment_list.indices(Goniometer())) == 0
    assert len(experiment_list.indices(Scan())) == 0


def test_experimentlist_models(experiment_list):
    # Get all the unique models
    b = experiment_list.beams()
    d = experiment_list.detectors()
    g = experiment_list.goniometers()
    s = experiment_list.scans()

    # Check we have the expected number
    assert len(b) == 3
    assert len(d) == 3
    assert len(g) == 3
    assert len(s) == 3

    # Check we have the expected order
    assert b[0] == experiment_list[0].beam
    assert b[1] == experiment_list[1].beam
    assert b[2] == experiment_list[2].beam

    assert d[0] == experiment_list[0].detector
    assert d[1] == experiment_list[1].detector
    assert d[2] == experiment_list[2].detector

    assert g[0] == experiment_list[0].goniometer
    assert g[0] == experiment_list[0].goniometer
    assert g[1] == experiment_list[1].goniometer

    assert s[2] == experiment_list[2].scan
    assert s[1] == experiment_list[1].scan
    assert s[2] == experiment_list[2].scan


def test_experimentlist_to_dict(experiment_list):
    # Convert the list to a dictionary
    obj = experiment_list.to_dict()

    # Check this is the right object
    assert obj["__id__"] == "ExperimentList"

    # Check length of items
    assert len(obj["experiment"]) == 5
    assert len(obj["beam"]) == 3
    assert len(obj["detector"]) == 3
    assert len(obj["goniometer"]) == 3
    assert len(obj["scan"]) == 3

    # The expected models
    b = [0, 1, 2, 1, 0]
    d = [0, 1, 2, 1, 0]
    g = [0, 1, 2, 1, 0]
    s = [0, 1, 2, 1, 0]

    # Check all the experiments
    for i, eobj in enumerate(obj["experiment"]):
        assert eobj["__id__"] == "Experiment"
        assert eobj["beam"] == b[i]
        assert eobj["detector"] == d[i]
        assert eobj["goniometer"] == g[i]
        assert eobj["scan"] == s[i]


def test_experimentlist_where(experiment_list):
    for beam in experiment_list.beams():
        assert beam is not None
        for i in experiment_list.where(beam=beam):
            assert experiment_list[i].beam is beam
    for goniometer in experiment_list.goniometers():
        assert goniometer is not None
        for i in experiment_list.where(goniometer=goniometer):
            assert experiment_list[i].goniometer is goniometer
    for scan in experiment_list.scans():
        assert scan is not None
        for i in experiment_list.where(scan=scan):
            assert experiment_list[i].scan is scan
    for detector in experiment_list.detectors():
        assert detector is not None
        for i in experiment_list.where(detector=detector):
            assert experiment_list[i].detector is detector


@pytest.fixture
def experiment_list():
    # Initialise a list of experiments
    experiments = ExperimentList()

    # Create a few beams
    b1 = Beam()
    b2 = Beam()
    b3 = Beam()

    # Create a few detectors
    d1 = Detector()
    d2 = Detector()
    d3 = Detector()

    # Create a few goniometers
    g1 = Goniometer()
    g2 = Goniometer()
    g3 = Goniometer()

    # Create a few scans
    s1 = Scan()
    s2 = Scan()
    s3 = Scan()

    # Create a list of models
    b = [b1, b2, b3, b2, b1]
    d = [d1, d2, d3, d2, d1]
    g = [g1, g2, g3, g2, g1]
    s = [s1, s2, s3, s2, s1]
    ident = ["sausage", "eggs", "bacon", "toast", "beans"]

    # Populate with various experiments
    for i in range(5):
        experiments.append(
            Experiment(
                beam=b[i],
                detector=d[i],
                goniometer=g[i],
                scan=s[i],
                identifier=ident[i],
            )
        )

    return experiments


def test_experimentlist_factory_from_json(monkeypatch, dials_data):
    data_dir = dials_data("experiment_test_data")
    dials_data_root = data_dir / ".."
    # Get all the filenames
    filename1 = str(data_dir / "experiment_1.json")
    filename3 = str(data_dir / "experiment_3.json")
    filename4 = str(data_dir / "experiment_4.json")

    with monkeypatch.context() as m:
        m.setenv("DIALS_DATA", str(dials_data_root.resolve()))
        el1 = ExperimentListFactory.from_json_file(filename1)
        el3 = ExperimentListFactory.from_json_file(filename3)
        el4 = ExperimentListFactory.from_json_file(filename4)

    # All the experiment lists should be the same length
    assert len(el1) == 1
    assert len(el1) == len(el3)
    assert len(el1) == len(el4)

    # Check all the models are the same
    for e in zip(el1, el3, el4):
        e1 = e[0]
        assert e1.imageset
        assert e1.beam
        assert e1.detector
        assert e1.goniometer
        assert e1.scan
        assert e1.crystal
        for ee in e[1:]:
            assert e1.imageset == ee.imageset
            assert e1.beam == ee.beam
            assert e1.detector == ee.detector
            assert e1.goniometer == ee.goniometer
            assert e1.scan == ee.scan
            assert e1.crystal == ee.crystal


def test_experimentlist_factory_from_pickle(monkeypatch, dials_data):
    data_dir = dials_data("experiment_test_data")
    dials_data_root = data_dir / ".."
    # Get all the filenames
    filename1 = str(data_dir / "experiment_1.json")

    # Read all the experiment lists in
    with monkeypatch.context() as m:
        m.setenv("DIALS_DATA", str(dials_data_root.resolve()))
        el1 = ExperimentListFactory.from_json_file(filename1)

    # Pickle then load again
    el2 = pickle.loads(pickle.dumps(el1))

    # All the experiment lists should be the same length
    assert len(el1) == 1
    assert len(el1) == len(el2)

    # Check all the models are the same
    for e1, e2 in zip(el1, el2):
        assert e1.imageset and e1.imageset == e2.imageset
        assert e1.beam and e1.beam == e2.beam
        assert e1.detector and e1.detector == e2.detector
        assert e1.goniometer and e1.goniometer == e2.goniometer
        assert e1.scan and e1.scan == e2.scan
        assert e1.crystal and e1.crystal == e2.crystal


def test_experimentlist_factory_from_args(monkeypatch, dials_data):
    pytest.importorskip("dials")

    data_dir = dials_data("experiment_test_data")
    dials_data_root = data_dir / ".."

    # Get all the filenames
    filenames = [
        str(data_dir / "experiment_1.json"),
        str(data_dir / "experiment_3.json"),
        str(data_dir / "experiment_4.json"),
    ]

    # Get the experiments from a list of filenames
    with monkeypatch.context() as m:
        m.setenv("DIALS_DATA", str(dials_data_root.resolve()))
        experiments = ExperimentListFactory.from_args(filenames)

    assert len(experiments) == 3
    for experiment in experiments:
        assert experiment.imageset
        assert experiment.beam
        assert experiment.detector
        assert experiment.goniometer
        assert experiment.scan


def test_experimentlist_factory_from_imageset():
    imageset = Format.get_imageset(["filename.cbf"], as_imageset=True)
    imageset.set_beam(Beam(), 0)
    imageset.set_detector(Detector(), 0)

    crystal = Crystal((1, 0, 0), (0, 1, 0), (0, 0, 1), space_group_symbol="P1")

    experiments = ExperimentListFactory.from_imageset_and_crystal(imageset, crystal)

    assert len(experiments) == 1
    assert experiments[0].imageset
    assert experiments[0].beam
    assert experiments[0].detector is not None
    assert experiments[0].crystal


def test_experimentlist_factory_from_sequence():
    filenames = ["filename_%01d.cbf" % (i + 1) for i in range(0, 2)]

    imageset = Format.get_imageset(
        filenames,
        beam=Beam(),
        detector=Detector(),
        goniometer=Goniometer(),
        scan=Scan((1, 2), (0, 1)),
        as_sequence=True,
    )

    crystal = Crystal((1, 0, 0), (0, 1, 0), (0, 0, 1), space_group_symbol="P1")

    experiments = ExperimentListFactory.from_imageset_and_crystal(imageset, crystal)

    assert len(experiments) == 1
    assert experiments[0].imageset
    assert experiments[0].beam
    assert experiments[0].detector is not None
    assert experiments[0].goniometer
    assert experiments[0].scan
    assert experiments[0].crystal


def test_experimentlist_dumper_dump_formats(monkeypatch, dials_data, tmp_path):
    data_dir = dials_data("experiment_test_data")
    dials_data_root = data_dir / ".."

    # Get all the filenames
    filename1 = str(data_dir / "experiment_1.json")

    # Read all the experiment lists in
    with monkeypatch.context() as m:
        m.setenv("DIALS_DATA", str(dials_data_root.resolve()))
        elist1 = ExperimentListFactory.from_json_file(filename1)

    # Dump as JSON file and reload
    filename = tmp_path / "temp1.json"
    elist1.as_json(filename)
    elist2 = ExperimentListFactory.from_json_file(filename)
    check(elist1, elist2)

    # Dump as split JSON file and reload
    filename = tmp_path / "temp2.json"
    elist1.as_json(filename, split=True)
    elist2 = ExperimentListFactory.from_json_file(filename)
    check(elist1, elist2)


def test_experimentlist_dumper_dump_scan_varying(monkeypatch, dials_data, tmp_path):
    data_dir = dials_data("experiment_test_data")
    dials_data_root = data_dir / ".."
    # Get all the filenames
    filename1 = str(data_dir / "experiment_1.json")

    # Read the experiment list in
    with monkeypatch.context() as m:
        m.setenv("DIALS_DATA", str(dials_data_root.resolve()))
        elist1 = ExperimentListFactory.from_json_file(filename1)

    # Make trivial scan-varying models
    crystal = elist1[0].crystal
    beam = elist1[0].beam
    goniometer = elist1[0].goniometer
    crystal.set_A_at_scan_points([crystal.get_A()] * 5)

    cov_B = flex.double([1e-5] * 9 * 9)
    crystal.set_B_covariance(cov_B)
    cov_B.reshape(flex.grid(1, 9, 9))
    cov_B_array = flex.double(flex.grid(5, 9, 9))
    for i in range(5):
        cov_B_array[i : (i + 1), :, :] = cov_B
    crystal.set_B_covariance_at_scan_points(cov_B_array)

    beam.set_s0_at_scan_points([beam.get_s0()] * 5)
    goniometer.set_setting_rotation_at_scan_points(
        [goniometer.get_setting_rotation()] * 5
    )

    # Dump as JSON file and reload
    filename = tmp_path / "temp.json"
    elist1.as_json(filename)
    elist2 = ExperimentListFactory.from_json_file(filename)
    check(elist1, elist2)


def test_experimentlist_dumper_dump_empty_sequence(tmp_path):
    filenames = [tmp_path / f"filename_{i}.cbf" for i in range(1, 3)]

    imageset = Format.get_imageset(
        filenames,
        beam=Beam((1, 0, 0)),
        detector=Detector(),
        goniometer=Goniometer(),
        scan=Scan((1, 2), (0.0, 1.0)),
        as_sequence=True,
    )

    crystal = Crystal((1, 0, 0), (0, 1, 0), (0, 0, 1), space_group_symbol="P1")

    experiments = ExperimentListFactory.from_imageset_and_crystal(imageset, crystal)

    filename = tmp_path / "temp.json"
    experiments.as_json(filename)
    experiments2 = ExperimentListFactory.from_json_file(filename, check_format=False)
    check(experiments, experiments2)


def test_experimentlist_dumper_dump_with_lookup(dials_data, tmp_path):
    data_dir = dials_data("centroid_test_data")

    # Copy to the tmp directory, because we need to unpack some files
    filename = shutil.copy(data_dir / "experiments_with_lookup.json", tmp_path)
    gain_bz2 = shutil.copy(data_dir / "lookup_gain.pickle.bz2", tmp_path)
    pedestal_bz2 = shutil.copy(data_dir / "lookup_pedestal.pickle.bz2", tmp_path)
    shutil.copy(data_dir / "lookup_mask.pickle", tmp_path)
    for image in data_dir.glob("centroid_000*.cbf"):
        shutil.copy(image, tmp_path)

    for f in [gain_bz2, pedestal_bz2]:
        with bz2.BZ2File(f) as compr:
            with open(f[:-4], "wb") as decompr:
                shutil.copyfileobj(compr, decompr)

    experiments = ExperimentListFactory.from_json_file(filename, check_format=True)

    imageset = experiments[0].imageset
    assert not imageset.external_lookup.mask.data.empty()
    assert not imageset.external_lookup.gain.data.empty()
    assert not imageset.external_lookup.pedestal.data.empty()
    assert imageset.external_lookup.mask.filename is not None
    assert imageset.external_lookup.gain.filename is not None
    assert imageset.external_lookup.pedestal.filename is not None
    assert imageset.external_lookup.mask.data.tile(0).data().all_eq(True)
    assert imageset.external_lookup.gain.data.tile(0).data().all_eq(1)
    assert imageset.external_lookup.pedestal.data.tile(0).data().all_eq(0)

    filename = tmp_path / "temp.json"
    experiments.as_json(filename)

    experiments = ExperimentListFactory.from_json_file(filename, check_format=True)

    imageset = experiments[0].imageset
    assert not imageset.external_lookup.mask.data.empty()
    assert not imageset.external_lookup.gain.data.empty()
    assert not imageset.external_lookup.pedestal.data.empty()
    assert imageset.external_lookup.mask.filename is not None
    assert imageset.external_lookup.gain.filename is not None
    assert imageset.external_lookup.pedestal.filename is not None
    assert imageset.external_lookup.mask.data.tile(0).data().all_eq(True)
    assert imageset.external_lookup.gain.data.tile(0).data().all_eq(1)
    assert imageset.external_lookup.pedestal.data.tile(0).data().all_eq(0)


def test_experimentlist_dumper_dump_with_bad_lookup(dials_data, tmpdir):
    filename = dials_data("centroid_test_data") / "experiments_with_bad_lookup.json"
    experiments = ExperimentListFactory.from_json_file(filename, check_format=False)

    imageset = experiments[0].imageset
    assert imageset.external_lookup.mask.data.empty()
    assert imageset.external_lookup.gain.data.empty()
    assert imageset.external_lookup.pedestal.data.empty()
    assert imageset.external_lookup.mask.filename is not None
    assert imageset.external_lookup.gain.filename is not None
    assert imageset.external_lookup.pedestal.filename is not None

    filename = tmpdir / "temp.json"
    experiments.as_json(filename)

    experiments = ExperimentListFactory.from_json_file(filename, check_format=False)

    imageset = experiments[0].imageset
    assert imageset.external_lookup.mask.data.empty()
    assert imageset.external_lookup.gain.data.empty()
    assert imageset.external_lookup.pedestal.data.empty()
    assert imageset.external_lookup.mask.filename is not None
    assert imageset.external_lookup.gain.filename is not None
    assert imageset.external_lookup.pedestal.filename is not None


def test_experimentlist_with_identifiers():
    # Initialise a list of experiments
    experiments = ExperimentList()

    experiments.append(
        Experiment(beam=Beam(s0=(0, 0, -1)), detector=Detector(), identifier="bacon")
    )

    experiments.append(
        Experiment(beam=Beam(s0=(0, 0, -1)), detector=Detector(), identifier="sausage")
    )

    with pytest.raises(Exception):
        experiments.append(
            Experiment(beam=Beam(), detector=Detector(), identifier="bacon")
        )

    d = experiments.to_dict()
    e2 = ExperimentListDict(d).decode()

    assert experiments[0].identifier == e2[0].identifier
    assert experiments[1].identifier == e2[1].identifier

    assert tuple(experiments.identifiers()) == ("bacon", "sausage")
    experiments[0].identifier = "spam"
    assert tuple(experiments.identifiers()) == ("spam", "sausage")

    experiments.append(Experiment(identifier="bacon"))
    experiments.select_on_experiment_identifiers(["spam", "bacon"])
    assert list(experiments.identifiers()) == ["spam", "bacon"]
    experiments.append(Experiment(identifier="ham"))
    experiments.append(Experiment(identifier="jam"))
    experiments.remove_on_experiment_identifiers(["spam", "jam"])
    assert list(experiments.identifiers()) == ["bacon", "ham"]


def test_load_models(dials_data):
    pytest.importorskip("h5py")
    filename = dials_data("image_examples") / "SACLA-MPCCD-run266702-0-subset.h5"

    # Test different ways of loading the data
    waves1, waves2, waves3, waves4, waves5 = [], [], [], [], []
    oris1, oris2, oris3, oris4, oris5 = [], [], [], [], []

    img = dxtbx.load(filename)

    # Test using dxtbx directly
    for i in range(img.get_num_images()):
        waves1.append(img.get_beam(i).get_wavelength())
        oris1.append(img.get_detector(i)[0].get_origin())

    # Test using the imageset clases
    imageset = img.get_imageset(str(filename))
    for i in range(len(imageset)):
        waves2.append(imageset.get_beam(i).get_wavelength())
        oris2.append(imageset.get_detector(i)[0].get_origin())

    # Test using imageset subsets
    imageset = img.get_imageset(str(filename))
    for i in range(len(imageset)):
        subset = imageset[i : i + 1]
        waves3.append(subset.get_beam(0).get_wavelength())
        oris3.append(subset.get_detector(0)[0].get_origin())

    # Test using pre-loaded experiments
    experiments = ExperimentListFactory.from_filenames([filename])
    for experiment in experiments:
        waves4.append(experiment.beam.get_wavelength())
        oris4.append(experiment.detector[0].get_origin())

    # Test using post-loaded experiments
    experiments = ExperimentListFactory.from_filenames([filename], load_models=False)
    for experiment in experiments:
        assert experiment.beam is None and experiment.detector is None
        experiment.load_models()
        waves5.append(experiment.beam.get_wavelength())
        oris5.append(experiment.detector[0].get_origin())

    for w1, w2, w3, w4, w5 in zip(waves1, waves2, waves3, waves4, waves5):
        assert w1 == w2 == w3 == w4 == w5

    for o1, o2, o3, o4, o5 in zip(oris1, oris2, oris3, oris4, oris5):
        assert o1 == o2 == o3 == o4 == o5


def test_partial_missing_model_serialization():
    goniometer = Goniometer()
    elist = ExperimentList([Experiment(), Experiment(goniometer=goniometer)])
    elist_ = ExperimentListFactory.from_dict(elist.to_dict())
    check(elist, elist_)


def test_experiment_type():
    experiment = Experiment()
    assert experiment.get_type() == ExperimentType.STILL
    experiment.goniometer = Goniometer()
    assert experiment.get_type() == ExperimentType.STILL
    experiment.scan = Scan()
    assert experiment.get_type() == ExperimentType.STILL
    experiment.scan = Scan((1, 1000), (0, 0.05))
    assert experiment.get_type() == ExperimentType.ROTATION
    # Specifically test the bug from dxtbx#4 triggered by ending on 0°
    experiment.scan = Scan((1, 1800), (-90, 0.05))
    assert experiment.get_type() == ExperimentType.ROTATION

    experiment.beam = BeamFactory.make_polychromatic_beam(
        direction=(0, 0, -1),
        sample_to_source_distance=(100),
        probe=Probe.xray,
        wavelength_range=(1, 10),
    )

    assert experiment.get_type() == ExperimentType.LAUE

    experiment.scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"time_of_flight": list(range(10))}
    )
    assert experiment.get_type() == ExperimentType.TOF

    experiment.beam = Beam()
    experiment.scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"other_property": list(range(10))}
    )
    assert experiment.get_type() == ExperimentType.STILL


def check(el1, el2):
    # All the experiment lists should be the same length
    assert len(el1) == len(el2)

    # Check all the models are the same
    for e1, e2 in zip(el1, el2):
        assert compare_experiment(e1, e2)


def compare_experiment(exp1, exp2):
    return (
        exp1.imageset == exp2.imageset
        and exp1.beam == exp2.beam
        and exp1.detector == exp2.detector
        and exp1.goniometer == exp2.goniometer
        and exp1.scan == exp2.scan
        and exp1.profile == exp2.profile
        and exp1.scaling_model == exp2.scaling_model
        and exp1.identifier == exp2.identifier
    )


def test_experimentlist_from_file(dials_data, monkeypatch, tmpdir):
    # With the default check_format=True this file should fail to load with an
    # appropriate error as we can't find the images on disk
    data_dir = dials_data("experiment_test_data")
    dials_data_root = data_dir / ".."

    with monkeypatch.context() as m:
        m.delenv("DIALS_DATA", raising=False)
        with pytest.raises(IOError) as e:
            exp_list = ExperimentList.from_file(str(data_dir / "experiment_1.json"))
    assert e.value.errno == errno.ENOENT
    assert "No such file or directory" in str(e.value)
    assert "centroid_0001.cbf" in str(e.value)

    # Setting check_format=False should allow the file to load
    exp_list = ExperimentList.from_file(
        str(data_dir / "experiment_1.json"),
        check_format=False,
    )
    assert len(exp_list) == 1
    assert exp_list[0].beam

    # This allows expansion of environment variables in regression files, enabling the
    # file to load with check_format=True

    with monkeypatch.context() as m:
        m.setenv("DIALS_DATA", str(dials_data_root.resolve()))
        exp_list = ExperimentList.from_file(str(data_dir / "experiment_1.json"))
    assert len(exp_list) == 1
    assert exp_list[0].beam


def test_experimentlist_imagesequence_stills(dials_data):
    filenames = [
        str(dials_data("thaumatin_grid_scan") / f"thau_3_2_{i:04d}.cbf.bz2")
        for i in range(1, 4)
    ]
    experiments = ExperimentListFactory.from_filenames(filenames)

    assert len(experiments) == 3
    assert len(experiments.imagesets()) == 1

    # Convert experiment list to dict
    d = experiments.to_dict()

    # Decode the dict to get a new experiment list
    experiments2 = ExperimentListDict(d).decode()

    # Verify that this experiment is as we expect
    assert len(experiments2) == 3
    assert len(experiments2.imagesets()) == 1
    assert len(experiments2.goniometers()) == 1
    assert len(experiments2.detectors()) == 1
    assert len(experiments2.beams()) == 1
    assert len(experiments2.scans()) == 3
    for expt in experiments2:
        assert expt.imageset is experiments2.imagesets()[0]


def test_experimentlist_imagesequence_decode(mocker):
    # These models are shared between experiments
    beam = Beam(s0=(0, 0, -1))
    detector = Detector()
    gonio = Goniometer()

    # Construct the experiment list
    experiments = ExperimentList()
    for i in range(3):
        experiments.append(
            Experiment(
                beam=beam,
                detector=detector,
                scan=ScanFactory.make_scan(
                    image_range=(i + 1, i + 1),
                    exposure_times=[1],
                    oscillation=(0, 0),
                    epochs=[0],
                ),
                goniometer=gonio,
            )
        )

    # Convert experiment list to dict and manually insert a shared imageset
    d = experiments.to_dict()
    d["imageset"].append(
        {"__id__": "ImageSequence", "template": "Puck3_10_1_####.cbf.gz"}
    )
    for e in d["experiment"]:
        e["imageset"] = 0

    # Monkeypatch this function as we don't actually have an imageset
    make_sequence = mocker.patch.object(ExperimentListDict, "_make_sequence")
    # Ensure that if make_sequence is called more than once it returns a different
    # value each time
    make_sequence.side_effect = lambda *args, **kwargs: mocker.MagicMock()

    # Decode the dict to get a new experiment list
    experiments2 = ExperimentListDict(d).decode()

    # This function should only be called once per imageset
    make_sequence.assert_called_once()

    # Verify that this experiment is as we expect
    assert len(experiments2) == 3
    assert len(experiments2.imagesets()) == 1
    assert len(experiments2.goniometers()) == 1
    assert len(experiments2.detectors()) == 1
    assert len(experiments2.beams()) == 1
    assert len(experiments2.scans()) == 3
    for expt in experiments2:
        assert expt.imageset is experiments2.imagesets()[0]


def test_experimentlist_change_basis(dials_data):
    experiments = ExperimentList()
    for i in range(4):
        experiments.extend(
            ExperimentList.from_file(
                dials_data("vmxi_proteinase_k_sweeps") / ("experiments_%i.expt" % i),
                check_format=False,
            )
        )
    reindexed_uc = (68.368, 103.968, 68.368, 90.000, 90.000, 90.000)
    reindexed_sg = sgtbx.space_group_info("P 4 2 2 (b,c,a)").group()
    cb_op = sgtbx.change_of_basis_op("-a,-c,-b")
    for cb_op in (cb_op, [cb_op] * len(experiments)):
        expts_rdx = experiments.change_basis(cb_op)
        for expt in expts_rdx:
            assert expt.crystal.get_unit_cell().parameters() == pytest.approx(
                reindexed_uc, abs=0.1
            )
            assert expt.crystal.get_space_group() == reindexed_sg

    experiments.change_basis(cb_op, in_place=True)
    for expt in experiments:
        assert expt.crystal.get_unit_cell().parameters() == pytest.approx(
            reindexed_uc, abs=0.1
        )
        assert expt.crystal.get_space_group() == reindexed_sg

    with pytest.raises(AssertionError):
        experiments.change_basis([cb_op, cb_op])


def test_path_iterator(monkeypatch):
    """Test the pathname iterator that avoids excessive file calls"""

    @classmethod
    def _fake_open_file(cls, name):
        """Mock replacement for Format's open_file"""
        if name in ("a", "b", os.path.join("dir", "c"), os.path.join("dir", "d"), "e"):
            return mock.Mock()
        elif name.startswith("dir"):
            err = OSError()
            err.errno = errno.EISDIR
            # raise IOError(errno=errno.EISDIR)
            raise err
        assert False

    # Path the lookup of files
    listdir = mock.Mock(return_value=["c", "dir2", "d"])
    monkeypatch.setattr(os, "listdir", listdir)
    # Replace Format.open_file with a tame version
    monkeypatch.setattr(Format, "open_file", _fake_open_file)

    it = dxtbx.model.experiment_list._openingpathiterator(["a", "b", "dir", "e"])
    assert list(it) == [
        "a",
        "b",
        os.path.join("dir", "c"),
        os.path.join("dir", "d"),
        "e",
    ]
    listdir.assert_called_once_with("dir")

    # Test that the list is sorted
    it = dxtbx.model.experiment_list._openingpathiterator(["e", "a", "b", "dir"])
    assert list(it) == [
        "a",
        "b",
        os.path.join("dir", "c"),
        os.path.join("dir", "d"),
        "e",
    ]


def test_extract_metadata_record():
    """Make sure we can read a metadataobject from a format instance"""
    fmt = mock.MagicMock()
    fmt.get_image_file.return_value = "filename_000.cbf"
    fmt.get_scan.return_value = None
    record = dxtbx.model.experiment_list.ImageMetadataRecord.from_format(fmt)
    assert record.beam is fmt.get_beam()
    assert record.detector is fmt.get_detector()
    assert record.goniometer is fmt.get_goniometer()
    assert record.scan is None
    assert record.index is None


def _equal_but_not_same(thing):
    object_1 = (thing,)
    object_2 = (thing,)
    assert object_1 == object_2
    assert object_1 is not object_2
    return object_1, object_2


def test_merge_metadata_record():
    """Test that merging metadata records works correctly"""
    beam_a, beam_b = _equal_but_not_same("beam")
    detector_a, detector_b = _equal_but_not_same("detector")
    gonio_a, gonio_b = _equal_but_not_same("goniometer")

    a = dxtbx.model.experiment_list.ImageMetadataRecord(
        beam=beam_a, detector=detector_a, goniometer=gonio_a
    )
    b = dxtbx.model.experiment_list.ImageMetadataRecord(
        beam=beam_b, detector=detector_b, goniometer=gonio_b
    )
    pre_hash = hash(a)
    assert a.beam is not b.beam
    assert a.detector is not b.detector
    assert a.goniometer is not b.goniometer
    # This should do something
    assert b.merge_metadata_from(a)
    assert hash(a) == pre_hash, "a changed after merge"
    # Make sure metadata was merged
    assert a.beam is b.beam
    assert a.detector is b.detector
    assert a.goniometer is b.goniometer
    # This should NOT do something
    assert not a.merge_metadata_from(a)
    assert hash(a) == pre_hash


def test_merge_all_metadata():
    """Test that merging metadata over a whole list of records works"""
    beam_a, beam_b = _equal_but_not_same("beam")
    gonio_a, gonio_b = _equal_but_not_same("goniometer")

    a = dxtbx.model.experiment_list.ImageMetadataRecord(
        beam=beam_a, detector=object(), goniometer=gonio_a
    )
    b = dxtbx.model.experiment_list.ImageMetadataRecord(
        beam=beam_b, detector=object(), goniometer=gonio_b
    )
    records = [a, b]
    dxtbx.model.experiment_list._merge_model_metadata(records)
    assert a.beam is b.beam
    assert a.goniometer is b.goniometer
    assert a.detector is not b.detector


def test_merge_scan():
    """Test merging logic of scans"""
    scanA = mock.Mock(spec=Scan)
    scanB = mock.Mock(spec=Scan)
    recordA = mock.Mock(scan=scanA)
    recordB = mock.Mock(
        scan=scanB,
        beam=recordA.beam,
        detector=recordA.detector,
        goniometer=recordA.goniometer,
    )
    result = dxtbx.model.experiment_list._merge_scans([recordA, recordB])
    assert result == [recordA]
    scanA.append.assert_called_once_with(scanB)

    # Change some metadata in recordB so it doesn't match
    scanA.reset_mock()
    recordB.beam = mock.Mock()
    assert dxtbx.model.experiment_list._merge_scans([recordA, recordB]) == [
        recordA,
        recordB,
    ]


def test_groupby_template_none():
    Fo = collections.namedtuple("Fo", ["template"])
    objs = [Fo(1), Fo(2), Fo(2), Fo(None), Fo(None), Fo("something")]
    result = list(dxtbx.model.experiment_list._groupby_template_is_none(objs))
    assert result == [
        [Fo(1)],
        [Fo(2)],
        [Fo(2)],
        [Fo(None), Fo(None)],
        [Fo("something")],
    ]


def test_iterate_with_previous():
    sample = list(range(5))
    assert list(dxtbx.model.experiment_list._iterate_with_previous(sample)) == [
        (None, 0),
        (0, 1),
        (1, 2),
        (2, 3),
        (3, 4),
    ]


def test_create_single_sequence(single_sequence_filenames):
    experiments = ExperimentListFactory.from_filenames(single_sequence_filenames)
    assert len(experiments) == 1
    imagesets = experiments.imagesets()
    assert imagesets[0].get_format_class()
    assert len(imagesets) == 1
    assert len(imagesets[0]) == 9


def test_create_multiple_sequences(multiple_sequence_filenames):
    experiments = ExperimentListFactory.from_filenames(multiple_sequence_filenames)
    assert len(experiments) == 2
    imagesets = experiments.imagesets()
    assert len(imagesets) == 2
    assert imagesets[0].get_format_class()
    assert imagesets[1].get_format_class()
    assert len(imagesets[0]) == 3
    assert len(imagesets[1]) == 3


def test_create_multiple_blocks(multiple_block_filenames):
    experiments = ExperimentListFactory.from_filenames(multiple_block_filenames)
    assert len(experiments) == 24
    imagesets = experiments.imagesets()
    assert len(imagesets) == 24
    assert [len(im) for im in imagesets] == [9] + [1] * 23


def test_from_null_sequence():
    filenames = ["template_%2d.cbf" % (i + 1) for i in range(0, 10)]
    sequence = Format.get_imageset(
        filenames,
        beam=Beam((0, 0, 1)),
        detector=Detector(),
        goniometer=Goniometer((1, 0, 0)),
        scan=Scan((1, 10), (0, 0.1)),
    )

    # Create the experiments
    experiments = ExperimentListFactory.from_sequence_and_crystal(
        sequence, crystal=None
    )
    assert len(experiments) == 1
    imagesets = experiments.imagesets()
    assert imagesets[0].get_format_class()
    assert len(imagesets) == 1
    assert imagesets[0].get_beam() == sequence.get_beam()
    assert imagesets[0].get_detector() == sequence.get_detector()
    assert imagesets[0].get_goniometer() == sequence.get_goniometer()
    assert imagesets[0].get_scan() == sequence.get_scan()


def test_from_templates(dials_data):
    template = dials_data("insulin") / "insulin_1_###.img"
    expts = ExperimentList.from_templates([template])
    assert len(expts) == 1
    assert expts[0].imageset.get_template() == str(template)
    assert len(expts[0].imageset) == 45


def test_experiment_list_all():
    experiments = ExperimentList()
    assert experiments.all_same_type()
    for i in range(3):
        experiments.append(Experiment())

    assert experiments.all_stills()
    assert experiments.all_same_type()
    experiments[0].goniometer = Goniometer()
    assert experiments.all_stills()
    assert experiments.all_same_type()
    experiments[1].goniometer = Goniometer()
    experiments[2].goniometer = Goniometer()
    assert experiments.all_stills()
    assert experiments.all_same_type()

    experiments[0].beam = BeamFactory.make_polychromatic_beam(
        direction=(0, 0, -1),
        sample_to_source_distance=(100),
        probe=Probe.xray,
        wavelength_range=(1, 10),
    )
    assert not experiments.all_stills()
    assert not experiments.all_same_type()
    experiments[1].beam = BeamFactory.make_polychromatic_beam(
        direction=(0, 0, -1),
        sample_to_source_distance=(100),
        probe=Probe.xray,
        wavelength_range=(1, 10),
    )
    experiments[2].beam = BeamFactory.make_polychromatic_beam(
        direction=(0, 0, -1),
        sample_to_source_distance=(100),
        probe=Probe.xray,
        wavelength_range=(1, 10),
    )
    assert experiments.all_laue()
    assert experiments.all_same_type()

    experiments[0].beam = Beam()
    assert not experiments.all_laue()
    assert not experiments.all_same_type()
    experiments[1].beam = Beam()
    experiments[2].beam = Beam()
    assert experiments.all_stills()
    assert experiments.all_same_type()

    experiments[0].scan = Scan((1, 1000), (0, 0.05))
    assert not experiments.all_stills()
    assert not experiments.all_same_type()
    experiments[1].scan = Scan((1, 1000), (0, 0.05))
    experiments[2].scan = Scan((1, 1000), (0, 0.05))
    assert experiments.all_rotations()
    assert experiments.all_same_type()

    experiments[0].scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"time_of_flight": list(range(10))}
    )
    assert not experiments.all_rotations()
    assert not experiments.all_same_type()
    experiments[1].scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"time_of_flight": list(range(10))}
    )
    experiments[2].scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"time_of_flight": list(range(10))}
    )
    assert experiments.all_tof()
    assert experiments.all_same_type()

    experiments[0].scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"other_property": list(range(10))}
    )
    assert not experiments.all_tof()
    assert not experiments.all_same_type()
    experiments[1].scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"other_property": list(range(10))}
    )
    experiments[2].scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"other_property": list(range(10))}
    )
    assert experiments.all_stills()
    assert experiments.all_same_type()


def test_history(tmp_path):
    # Check basic construction and pickling
    history = History(["foo"])
    assert history.get_history() == ["foo"]
    history2 = pickle.loads(pickle.dumps(history))
    assert history2.get_history() == history.get_history()

    # Check append history, which adds a timestamp
    history = History()
    history.append_history_item("dxtbx.program", "v42.0.0", "flag")
    h = history.get_history()
    assert len(h) == 1
    timestamp, program, version, flag = h[-1].split("|")
    assert dateutil.parser.isoparse(timestamp)
    assert program == "dxtbx.program"
    assert version == "v42.0.0"
    assert flag == "flag"

    # Now check saving and loading, which adds new history items
    el = ExperimentList()
    el.append(Experiment())
    el[0].history = history
    el.as_file(tmp_path / "temp.expt")
    el2 = ExperimentList.from_file(tmp_path / "temp.expt")
    assert el2[0].history.get_history()[0:2] == el[0].history.get_history()

    _, program, _ = el2[0].history.get_history()[-1].split("|")
    # This module called as_json, so its name is in the history
    assert program == __name__

    # Check flags
    el.as_file(tmp_path / "temp2.expt", history_as_integrated=True)
    check = ExperimentList.from_file(tmp_path / "temp2.expt")[0].history.get_history()[
        -1
    ]
    assert "integrated" in check
    el.as_file(tmp_path / "temp3.expt", history_as_scaled=True)
    check = ExperimentList.from_file(tmp_path / "temp3.expt")[0].history.get_history()[
        -1
    ]
    assert "scaled" in check
    el.as_file(
        tmp_path / "temp4.expt", history_as_integrated=True, history_as_scaled=True
    )
    check = ExperimentList.from_file(tmp_path / "temp4.expt")[0].history.get_history()[
        -1
    ]
    assert "integrated,scaled" in check



from __future__ import annotations

import os
import re
import sys
import xml.etree.ElementTree as ET

import numpy as np

sys.path.append(os.path.join(os.getcwd(), "build"))

try:
    import pyterse
except ImportError:
    pyterse = None


from dxtbx import flumpy
from dxtbx.format.Format import Format
from dxtbx.model import ScanFactory
from dxtbx.model.beam import Probe


class FormatTRPX(Format):
    @staticmethod
    def understand(image_file):
        try:
            with FormatTRPX.open_file(image_file, "rb") as fh:
                header = fh.read(24)

        except OSError:
            return False

        # Check if header starts with '<Terse'
        is_trpx = header.startswith(b"<Terse")

        return is_trpx

    @staticmethod
    def _read_metadata(image_file):
        hd = {}
        with FormatTRPX.open_file(image_file, "rb") as f:
            header_data = b""
            while True:
                chunk = f.read(1024)
                if not chunk:
                    # End of file reached without finding '/>'
                    raise RuntimeError("Failed to find the end of the header.")
                header_data += chunk
                if b"/>" in header_data:
                    end_index = header_data.find(b"/>") + 2
                    header_data = header_data[:end_index]
                    break

            try:
                header_str = header_data.decode("utf-8")
            except UnicodeDecodeError as e:
                raise RuntimeError(f"Failed to decode header data as UTF-8: {e}")

            try:
                root = ET.fromstring(header_str)
            except ET.ParseError as e:
                raise RuntimeError(f"Failed to parse XML header: {e}")

            hd["prolix_bits"] = int(root.attrib.get("prolix_bits", "12"))
            hd["signed"] = int(root.attrib.get("signed", "0"))
            hd["block"] = int(root.attrib.get("block", "12"))
            hd["memory_size"] = int(root.attrib.get("memory_size", "0"))
            hd["number_of_values"] = int(root.attrib.get("number_of_values", "0"))
            hd["number_of_frames"] = int(root.attrib.get("number_of_frames", "1"))

            hd["pixel_size"] = (0.055, 0.055)
            hd["trusted_range"] = (0, 65535)
            hd["distance"] = 478.0
            hd["start_tilt"] = 0.0
            hd["delta_tilt"] = 0.0
            hd["exposure"] = 1.0

            if "dimensions" in root.attrib:
                dimensions_str = root.attrib["dimensions"]
                dimensions = list(map(int, dimensions_str.strip().split()))
                hd["image_size"] = dimensions
            else:
                num_values = hd["number_of_values"]
                image_dimension = int(np.sqrt(num_values))
                hd["image_size"] = (image_dimension, image_dimension)

            hd["distance"] = float(root.attrib.get("distance", "478.0"))
            pixel_size_str = root.attrib.get("pixel_size", "0.055 0.055")
            hd["pixel_size"] = tuple(map(float, pixel_size_str.strip().split()))
            trusted_range_str = root.attrib.get("trusted_range", "0 65535")
            hd["trusted_range"] = tuple(map(float, trusted_range_str.strip().split()))

        return hd

    def _start(self):
        """Open the image file and read useful metadata into an internal dictionary"""
        self._header_dictionary = self._read_metadata(self._image_file)

    def _goniometer(self):
        """Dummy goniometer, 'vertical' as the images are viewed."""
        goniometer = self._goniometer_factory.known_axis((0, -1, 0))
        return goniometer

    def _detector(self):
        beam_centre = [
            (p * i) / 2
            for p, i in zip(
                self._header_dictionary["pixel_size"],
                self._header_dictionary["image_size"],
            )
        ]
        d = self._detector_factory.simple(
            "PAD",
            self._header_dictionary["distance"],
            beam_centre,
            "+x",
            "-y",
            self._header_dictionary["pixel_size"],
            self._header_dictionary["image_size"],
            self._header_dictionary["trusted_range"],
        )
        return d

    def _beam(self):
        """Unpolarized beam, default energy 200 keV"""
        beam = self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=0.02508,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )
        return beam

    def _scan(self):
        """Scan model for this image, filling out any unavailable items with dummy values"""
        alpha = self._header_dictionary.get("start_tilt", 0.0)
        dalpha = self._header_dictionary.get("delta_tilt", 1.0)
        exposure = self._header_dictionary.get("exposure", 0.0)
        oscillation = (alpha, dalpha)
        fname = os.path.split(self._image_file)[-1]
        s = fname.split("_")[-1].split(".")[0]
        try:
            index = int(re.match(".*?([0-9]+)$", s).group(1))
        except AttributeError:
            index = 1
        scan = ScanFactory.make_scan((index, index), exposure, oscillation, {index: 0})
        return scan

    def get_raw_data(self):
        if pyterse is None:
            raise ImportError(
                "The package pyterse is not installed. Please install it to read TRPX files."
            )
        terse = pyterse.Terse.load(self._image_file)
        decompressed_data = terse.prolix()
        raw_data_flex = flumpy.from_numpy(decompressed_data)
        return raw_data_flex



from __future__ import annotations

import importlib.metadata
import logging
import sys

# REMOVE and inline when Python 3.10 is minimum
if sys.version_info < (3, 10):
    profile_entry_points = importlib.metadata.entry_points().get(
        "dxtbx.profile_model", []
    )
else:
    profile_entry_points = importlib.metadata.entry_points(group="dxtbx.profile_model")


class ProfileModelFactory:
    """
    A factory to create a profile model
    """

    @staticmethod
    def from_dict(obj):
        """
        Given a dictionary, convert to a profile model
        """
        if obj is None:
            return None
        for entry_point in profile_entry_points:
            if entry_point.name == obj["__id__"]:
                return entry_point.load().from_dict(obj)
        logging.getLogger("dxtbx.model.profile").warn(
            "No profile class %s registered" % obj["__id__"]
        )
        print(
            "dxtbx.model.profile: WARNING: No profile class %s registered"
            % obj["__id__"]
        )
        return None



from __future__ import annotations

from collections.abc import Iterable

import natsort

import boost_adaptbx.boost.python

import dxtbx.format.image  # noqa: F401, import dependency for unpickling
import dxtbx.format.Registry
from dxtbx.sequence_filenames import group_files_by_imageset, template_image_range

try:
    from .dxtbx_imageset_ext import (
        ExternalLookup,
        ExternalLookupItemBool,
        ExternalLookupItemDouble,
        ImageGrid,
        ImageSequence,
        ImageSet,
        ImageSetData,
    )
except ModuleNotFoundError:
    from dxtbx_imageset_ext import (  # type: ignore
        ExternalLookup,
        ExternalLookupItemBool,
        ExternalLookupItemDouble,
        ImageGrid,
        ImageSequence,
        ImageSet,
        ImageSetData,
    )

ext = boost_adaptbx.boost.python.import_ext("dxtbx_ext")

__all__ = (
    "ExternalLookup",
    "ExternalLookupItemBool",
    "ExternalLookupItemDouble",
    "ImageGrid",
    "ImageSet",
    "ImageSetData",
    "ImageSetFactory",
    "ImageSetLazy",
    "ImageSequence",
    "MemReader",
)


def _expand_template_to_sorted_filenames(
    template: str, indices: Iterable[int]
) -> list[str]:
    """Expand a template string to a list of filenames.

    Args:
        template: The template string, with a block of '#' to replace
        indices:  The numeric indices to insert
    """
    pfx = template.split("#")[0]
    sfx = template.split("#")[-1]
    count = template.count("#")
    if count == 1:
        # Special handling for a template with a single "#", which does not
        # assume a zero-padded index.
        filenames = [f"{pfx}{index}{sfx}" for index in indices]
    else:
        filenames = [f"{pfx}{index:0{count}}{sfx}" for index in indices]
    return natsort.natsorted(filenames)


class MemReader:
    """A reader for data already loaded in memory"""

    def __init__(self, images):
        self._images = images

    def copy(self, paths):
        """
        Experimental implementation where a copy of the reader also copies all
        the data
        """
        return MemReader(self._images)

    def paths(self):
        return ["" for im in self._images]

    def identifiers(self):
        return self.paths()

    def __len__(self):
        return len(self._images)

    def read(self, index):
        format_instance = self._images[index]
        return format_instance.get_raw_data()

    @staticmethod
    def is_single_file_reader():
        return False

    @staticmethod
    def master_path():
        return ""


@boost_adaptbx.boost.python.inject_into(ImageSet)
class _:
    """
    A class to inject additional methods into the imageset class
    """

    def __getitem__(self, item):
        """Get an item from the image set stream.

        If the item is an index, read and return the image at the given index.
        Otherwise, if the item is a slice, then create a new ImageSet object
        with the given number of array indices from the slice.

        Params:
            item The index or slice

        Returns:
            An image or new ImageSet object
        """
        if isinstance(item, slice):
            start = item.start or 0
            stop = item.stop or len(self)
            if item.step is not None and item.step != 1:
                raise IndexError("Step must be 1")
            if self.data().has_single_file_reader():
                reader = self.reader().copy(self.reader().paths(), stop - start)
            else:
                reader = self.reader().copy(self.reader().paths())
            return self.partial_set(reader, start, stop)
        else:
            return self.get_corrected_data(item)

    def __iter__(self):
        """Iterate over the array indices and read each image in turn."""
        for i in range(len(self)):
            yield self[i]

    def get_vendortype(self, index):
        """Get the vendor information."""
        return self.data().get_vendor()

    def get_format_class(self):
        """Get format class name"""
        return self.data().get_format_class()

    def get_spectrum(self, index):
        """Get the spectrum if available"""
        kwargs = self.params()
        if self.data().has_single_file_reader():
            format_instance = self.get_format_class().get_instance(
                self.data().get_master_path(), **kwargs
            )
        else:
            format_instance = self.get_format_class().get_instance(
                self.get_path(index), **kwargs
            )
        try:
            return format_instance.get_spectrum(self.indices()[index])
        except TypeError:
            return format_instance.get_spectrum()

    def params(self):
        """Get the parameters"""
        return self.data().get_params()

    def get_detectorbase(self, index):
        """
        A function to be injected into the imageset to get the detectorbase instance
        """
        kwargs = self.params()
        if self.data().has_single_file_reader():
            format_instance = self.get_format_class().get_instance(
                self.data().get_master_path(), **kwargs
            )
            return format_instance.get_detectorbase(self.indices()[index])
        else:
            format_instance = self.get_format_class().get_instance(
                self.get_path(index), **kwargs
            )
            return format_instance.get_detectorbase()

    def reader(self):
        """
        Return the reader
        """
        return self.data().reader()

    def masker(self):
        """
        Return the masker
        """
        return self.data().masker()

    def paths(self):
        """
        Return the list of paths
        """
        if self.data().has_single_file_reader():
            return [self.get_path(i) for i in range(len(self))]
        else:
            return [self.reader().paths()[i] for i in self.indices()]


class ImageSetLazy(ImageSet):
    """
    Lazy ImageSet class that doesn't necessitate setting the models ahead of time.
    Only when a particular model (like detector or beam) for an image is requested,
    it sets the model using the format class and then returns the model
    """

    def _get_item_from_parent_or_format(self, item_name, index):
        """
        Obtain an $item_name (eg. detector, beam, ...) of the given index from
        the parent class using get_detector, get_beam, ...
        If the parent class returns None then lookup the item using the format
        class (if defined) and store a local reference to the item using
        self.set_detector, set_beam, ..
        """
        if index is None:
            index = 0
        item = getattr(super(), "get_" + item_name)(index)
        if item is None:
            # If check_format=False was used, then _current_instance_ will not be set, so assume a None is correct
            format_class = self.get_format_class()
            if (
                hasattr(format_class, "_current_instance_")
                and format_class._current_instance_ is not None
            ):
                format_instance = format_class._current_instance_
                getter_function = getattr(format_instance, "get_" + item_name)
                item = getter_function(self.indices()[index])
                setter_function = getattr(self, "set_" + item_name)
                setter_function(item, index)
        return item

    def get_detector(self, index=None):
        return self._get_item_from_parent_or_format("detector", index)

    def get_beam(self, index=None):
        return self._get_item_from_parent_or_format("beam", index)

    def get_mask(self, index=None):
        """
        ImageSet::get_mask internally dereferences a pointer to the _detector
        member of ImageSetData, so we ensure the detector gets populated first.
        """
        if getattr(super(), "get_detector")(index) is None:
            self._load_models(index)
        return self._get_item_from_parent_or_format("mask", index)

    def get_goniometer(self, index=None):
        return self._get_item_from_parent_or_format("goniometer", index)

    def get_scan(self, index=None):
        return self._get_item_from_parent_or_format("scan", index)

    def _load_models(self, index):
        if index is None:
            index = 0
        # Sets the list for detector, beam etc before being accessed by functions in imageset.h
        self.get_detector(index)
        self.get_beam(index)
        self.get_goniometer(index)
        self.get_scan(index)

    def __getitem__(self, item):
        if isinstance(item, slice):
            start = item.start or 0
            stop = item.stop or len(self)
            if item.step is not None and item.step != 1:
                raise IndexError("Step must be 1")
            if self.data().has_single_file_reader():
                reader = self.reader().copy(self.reader().paths(), stop - start)
            else:
                reader = self.reader().copy(self.reader().paths())
            return ImageSetLazy(
                self.data().partial_data(reader, start, stop),
                indices=self.indices()[item],
            )
        self._load_models(item)
        return super().__getitem__(item)

    def get_corrected_data(self, index):
        self._load_models(index)
        return super().get_corrected_data(index)

    def get_gain(self, index):
        self._load_models(index)
        return super().get_gain(index)


@boost_adaptbx.boost.python.inject_into(ImageSequence)
class _imagesequence:
    def __getitem__(self, item):
        """Get an item from the sequence stream.

        If the item is an index, read and return the image at the given index.
        Otherwise, if the item is a slice, then create a new Sequence object
        with the given number of array indices from the slice.

        Params:
            item The index or slice

        Returns:
            An image or new Sequence object

        """
        if not isinstance(item, slice):
            return self.get_corrected_data(item)
        else:
            if item.step is not None:
                raise IndexError("Sequences must be sequential")

            start = item.start or 0
            stop = item.stop or len(self)

            if self.data().has_single_file_reader():
                reader = self.reader().copy(self.reader().paths(), stop - start)
            else:
                reader = self.reader().copy(self.reader().paths())
            return self.partial_set(reader, start, stop)

    def get_template(self):
        """Return the template"""
        return self.data().get_template()


def _analyse_files(filenames):
    """Group images by filename into image sets.

    Params:
        filenames The list of filenames

    Returns:
        A list of (template, [indices], is_sequence)

    """
    # Analyse filenames to figure out how many imagesets we have
    filelist_per_imageset = group_files_by_imageset(filenames)

    def _indices_sequential_ge_zero(indices):
        """Determine if indices are sequential."""
        prev = indices[0]
        if prev < 0:
            return False
        for curr in indices[1:]:
            if curr != prev + 1:
                return False
            prev = curr

        return True

    def _is_imageset_a_sequence(template, indices):
        """Return True/False if the imageset is a sequence or not.

        Where more than 1 image that follow sequential numbers are given
        the images are catagorised as belonging to a sequence, otherwise they
        belong to an image set.

        """
        if len(indices) <= 1:
            return False
        indices = sorted(indices)
        return _indices_sequential_ge_zero(indices)

    # Label each group as either an imageset or a sequence.
    file_groups = []
    for template, indices in filelist_per_imageset.items():
        # Check if this imageset is a sequence
        is_sequence = _is_imageset_a_sequence(template, indices)

        # Append the items to the group list
        file_groups.append((template, indices, is_sequence))

    return file_groups


# FIXME Lots of duplication in this class, need to tidy up
class ImageSetFactory:
    """Factory to create imagesets and sequences."""

    @staticmethod
    def new(filenames, check_headers=False, ignore_unknown=False):
        """Create an imageset or sequence

        Params:
            filenames A list of filenames
            check_headers Check the headers to ensure all images are valid
            ignore_unknown Ignore unknown formats

        Returns:
            A list of imagesets

        """
        # Ensure we have enough images
        if isinstance(filenames, list):
            assert filenames
        elif isinstance(filenames, str):
            filenames = [filenames]
        else:
            raise RuntimeError("unknown argument passed to ImageSetFactory")

        # Analyse the filenames and group the images into imagesets.
        filelist_per_imageset = _analyse_files(filenames)

        # For each file list denoting an image set, create the imageset
        # and return as a list of imagesets. N.B sequences and image sets are
        # returned in the same list.
        imagesetlist = []
        for filelist in filelist_per_imageset:
            try:
                if filelist[2] is True:
                    iset = ImageSetFactory._create_sequence(filelist, check_headers)
                else:
                    iset = ImageSetFactory._create_imageset(filelist, check_headers)
                imagesetlist.append(iset)
            except Exception:
                if not ignore_unknown:
                    raise

        return imagesetlist

    @staticmethod
    def from_template(
        template,
        image_range=None,
        check_headers=False,
        check_format=True,
        beam=None,
        detector=None,
        goniometer=None,
        scan=None,
    ):
        """Create a new sequence from a template.

        Params:
            template The template argument
            image_range The image range
            check_headers Check the headers to ensure all images are valid

        Returns:
            A list of sequences

        """
        if not check_format:
            assert not check_headers

        # Import here as Format and Imageset have cyclic dependencies
        from dxtbx.format.Format import Format
        from dxtbx.format.FormatMultiImage import FormatMultiImage

        # Check the template is valid
        if "#" in template:
            # Get the template image range
            if image_range is None:
                image_range = template_image_range(template)

            # Set the image range
            indices = range(image_range[0], image_range[1] + 1)
            filenames = _expand_template_to_sorted_filenames(template, indices)
            if check_format:
                format_class = dxtbx.format.Registry.get_format_class_for_file(
                    filenames[0]
                )
            else:
                format_class = Format
        else:
            # Note, this assumes image stacks can only be written by dectris detectors,
            # but doesn't account for other imagestack formats like MRC.
            if "master" not in template:
                raise ValueError("Invalid template")
            filenames = [template]
            indices = range(image_range[0], image_range[1] + 1)
            if check_format:
                format_class = dxtbx.format.Registry.get_format_class_for_file(
                    filenames[0]
                )
            else:
                format_class = FormatMultiImage

        # Create the sequence object
        sequence = format_class.get_imageset(
            filenames,
            single_file_indices=indices,
            template=template,
            as_sequence=True,
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            check_format=check_format,
        )

        return [sequence]

    @staticmethod
    def _create_imageset(filelist, check_headers):
        """Create an image set"""
        # Extract info from filelist
        template, indices, is_sequence = filelist

        # Get the template format
        if "#" in template:
            filenames = _expand_template_to_sorted_filenames(template, indices)
        else:
            filenames = [template]

        # Get the format object
        format_class = dxtbx.format.Registry.get_format_class_for_file(filenames[0])

        # Create and return the imageset
        return format_class.get_imageset(filenames, as_imageset=True)

    @staticmethod
    def _create_sequence(filelist, check_headers):
        """Create a sequence"""
        template, indices, is_sequence = filelist

        # Expand the template if necessary
        if "#" in template:
            filenames = _expand_template_to_sorted_filenames(template, indices)
        else:
            filenames = [template]

        # Get the format object
        format_class = dxtbx.format.Registry.get_format_class_for_file(filenames[0])

        return format_class.get_imageset(filenames, template=template, as_sequence=True)

    @staticmethod
    def make_imageset(
        filenames,
        format_class=None,
        check_format=True,
        single_file_indices=None,
        format_kwargs=None,
    ):
        """Create an image set"""
        # Import here as Format and Imageset have cyclic dependencies
        from dxtbx.format.Format import Format

        # So does FormatMultiImage
        from dxtbx.format.FormatMultiImage import FormatMultiImage

        # Get the format object
        if format_class is None:
            if check_format:
                format_class = dxtbx.format.Registry.get_format_class_for_file(
                    filenames[0]
                )
            else:
                if single_file_indices is None or len(single_file_indices) == 0:
                    format_class = Format
                else:
                    format_class = FormatMultiImage

        return format_class.get_imageset(
            filenames,
            single_file_indices=single_file_indices,
            as_imageset=True,
            format_kwargs=format_kwargs,
            check_format=check_format,
        )

    @staticmethod
    def make_sequence(
        template,
        indices,
        format_class=None,
        beam=None,
        detector=None,
        goniometer=None,
        scan=None,
        check_format=True,
        format_kwargs=None,
    ):
        """Create a sequence"""
        indices = sorted(indices)

        # Import here as Format and Imageset have cyclic dependencies
        from dxtbx.format.Format import Format
        from dxtbx.format.FormatMultiImage import FormatMultiImage

        # Get the template format
        if "#" in template:
            filenames = _expand_template_to_sorted_filenames(template, indices)
            # Get the format object and reader
            if format_class is None:
                if check_format:
                    format_class = dxtbx.format.Registry.get_format_class_for_file(
                        filenames[0]
                    )
                else:
                    format_class = Format
        else:
            filenames = [template]
            if format_class is None:
                if check_format:
                    format_class = dxtbx.format.Registry.get_format_class_for_file(
                        filenames[0]
                    )
                else:
                    format_class = FormatMultiImage

        # Set the image range
        array_range = (min(indices) - 1, max(indices))
        if scan is not None:
            assert array_range == scan.get_array_range()
            scan.set_batch_offset(array_range[0])

        return format_class.get_imageset(
            filenames,
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            format_kwargs=format_kwargs,
            template=template,
            as_sequence=True,
            check_format=check_format,
            single_file_indices=list(range(*array_range)),
        )

    @staticmethod
    def imageset_from_anyset(imageset):
        """Create a new ImageSet object from an imageset object. Converts ImageSequence to ImageSet."""
        if isinstance(imageset, ImageSetLazy):
            return ImageSetLazy(imageset.data(), imageset.indices())
        elif isinstance(imageset, ImageSequence) or isinstance(imageset, ImageSet):
            return ImageSet(imageset.data(), imageset.indices())
        else:
            raise ValueError("Unrecognized imageset type: %s" % str(type(imageset)))



from __future__ import annotations

import boost_adaptbx.boost.python  # noqa # lgtm # Ensure boost python libraries are in memory

try:
    from ..dxtbx_format_image_ext import *  # isort:skip # noqa: F403
except ModuleNotFoundError:
    from dxtbx_format_image_ext import *  # type: ignore # isort:skip # noqa: F403


__all__ = (  # noqa: F405
    "ImageBool",
    "ImageBuffer",
    "ImageDouble",
    "ImageInt",
    "ImageTileBool",
    "ImageTileDouble",
    "ImageTileInt",
)



from __future__ import annotations

import os

import pytest

from dxtbx.format.FormatHDF5EigerNearlyNexus import FormatHDF5EigerNearlyNexus
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.model.goniometer import Goniometer

pytest.importorskip("h5py")


def test_biomax_vertical_goniometer_at_diamond(tmpdir):
    master_h5 = (
        "/dls/science/groups/scisoft/DIALS/"
        "regression-test-data/biomax-eiger-moche/tau1-tau_1_master.h5"
    )

    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test data not available")

    assert FormatHDF5EigerNearlyNexus.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatHDF5EigerNearlyNexus

    gonio = imageset.get_goniometer()

    assert isinstance(gonio, Goniometer)
    assert gonio.get_rotation_axis() == (0, 1, 0)


def test_semi_synthetic_dectris_eiger_nearly_nexus(dials_data, tmpdir):
    master_h5 = dials_data("image_examples") / "dectris_eiger_master.h5"

    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatHDF5EigerNearlyNexus.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatHDF5EigerNearlyNexus

    image = imageset.get_raw_data(0)[0]
    assert min(image) == 0
    assert max(image) == 65535

    detector = imageset.get_detector()
    gonio = imageset.get_goniometer()
    scan = imageset.get_scan()
    beam = imageset.get_beam()

    panel = detector[0]
    assert panel.get_pixel_size() == pytest.approx((0.075, 0.075))
    assert panel.get_image_size() == (3110, 3269)
    assert panel.get_trusted_range() == (0, 12440)
    assert panel.get_fast_axis() == (1, 0, 0)
    assert panel.get_slow_axis() == (0, -1, 0)
    assert panel.get_thickness() == pytest.approx(0.45)
    assert panel.get_mu() == pytest.approx(3.96764, abs=1e-4)
    assert panel.get_material() == "Si"
    assert panel.get_origin() == pytest.approx((-120.556, 118.982, -134.255), abs=1e-3)
    assert panel.get_distance() == pytest.approx(134.255)

    assert isinstance(gonio, Goniometer)
    assert gonio.get_rotation_axis() == (1, 0, 0)
    assert gonio.get_fixed_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert gonio.get_setting_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)

    assert scan.get_oscillation() == pytest.approx((0, 0))
    assert scan.get_image_range() == (1, 1)

    assert beam.get_wavelength() == pytest.approx(0.980112, abs=1e-5)
    assert beam.get_s0() == pytest.approx((0, 0, -1 / beam.get_wavelength()))


@pytest.mark.skip(reason="test needs a smaller dataset in dials-data")
def test_soleil_Proxima2A_zenodo_1443110_data03():
    # https://zenodo.org/record/1221344#.XEHr_5ynx2Q
    master_h5 = "/dls/mx-scratch/rjgildea/zenodo/1221344/200Hz/3_5_200Hz_1_master.h5"

    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatHDF5EigerNearlyNexus.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatHDF5EigerNearlyNexus

    detector = imageset.get_detector()
    gonio = imageset.get_goniometer()
    scan = imageset.get_scan()
    beam = imageset.get_beam()

    panel = detector[0]
    assert panel.get_pixel_size() == pytest.approx((0.075, 0.075))
    assert panel.get_image_size() == (3110, 3269)
    assert panel.get_trusted_range() == (-1, 12440)
    assert panel.get_fast_axis() == (1, 0, 0)
    assert panel.get_slow_axis() == (0, -1, 0)
    assert panel.get_thickness() == pytest.approx(0.45)
    assert panel.get_mu() == pytest.approx(3.96763)
    assert panel.get_material() == "Si"
    assert panel.get_origin() == pytest.approx((-120.556, 118.982, -134.255), abs=1e-3)
    assert panel.get_distance() == pytest.approx(134.255)

    assert isinstance(gonio, Goniometer)
    assert gonio.get_rotation_axis() == (1, 0, 0)
    assert gonio.get_fixed_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert gonio.get_setting_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)

    assert scan.get_oscillation() == pytest.approx((0, 0.5))
    assert scan.get_image_range() == (1, 800)

    assert beam.get_wavelength() == pytest.approx(0.980112, abs=1e-5)
    assert beam.get_s0() == pytest.approx((0, 0, -1 / beam.get_wavelength()))



"""
Reads a multi-tile CBF image, discovering it's detector geometery
automatically, and builds a hierarchy if present
"""

from __future__ import annotations

import struct
import sys

import pycbf

from libtbx.utils import Sorry
from scitbx.array_family import flex
from scitbx.matrix import col, sqr

from dxtbx.format.FormatCBFMultiTile import FormatCBFMultiTile, FormatCBFMultiTileStill
from dxtbx.model import Detector


class FormatCBFMultiTileHierarchy(FormatCBFMultiTile):
    """An image reading class multi-tile CBF files"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_widefile(image_file.encode(), pycbf.MSG_DIGEST)

        # check if multiple arrays
        if cbf_handle.count_elements() <= 1:
            return False

        # we need the optional column equipment_component to build a hierarchy
        try:
            cbf_handle.find_category(b"axis")
            cbf_handle.find_column(b"equipment_component")
        except Exception as e:
            if "CBF_NOTFOUND" in str(e):
                return False
            raise e
        return True

    def _get_change_of_basis(self, axis_id):
        """Get the 4x4 homogenous coordinate matrix for a given axis.  Assumes
        the cbf handle has been initialized
        @param axis_id axis name of basis to get"""
        cbf = self._get_cbf_handle()
        axis_type = cbf.get_axis_type(axis_id)

        offset = col(cbf.get_axis_offset(axis_id))
        vector = col(cbf.get_axis_vector(axis_id)).normalize()
        setting, increment = cbf.get_axis_setting(axis_id)

        # change of basis matrix in homologous coordinates
        cob = None

        if axis_type == b"rotation":
            r3 = vector.axis_and_angle_as_r3_rotation_matrix(
                setting + increment, deg=True
            )
            cob = sqr(
                (
                    r3[0],
                    r3[1],
                    r3[2],
                    offset[0],
                    r3[3],
                    r3[4],
                    r3[5],
                    offset[1],
                    r3[6],
                    r3[7],
                    r3[8],
                    offset[2],
                    0,
                    0,
                    0,
                    1,
                )
            )
        elif axis_type == b"translation":
            translation = offset + vector * (setting + increment)
            cob = sqr(
                (
                    1,
                    0,
                    0,
                    translation[0],
                    0,
                    1,
                    0,
                    translation[1],
                    0,
                    0,
                    1,
                    translation[2],
                    0,
                    0,
                    0,
                    1,
                )
            )
        else:
            raise Sorry("Unrecognized vector type: %s" % axis_type)

        return cob

    def _get_cumulative_change_of_basis(self, axis_id):
        """Get the 4x4 homogenous coordinate matrix for a given axis, combining it with the change of
        basis matrices of parent axes with the same equipment component as the given axis. Assumes
        the cbf handle has been initialized
        @param axis_id axis name of basis to get
        @return (parent, change of basis matrix), where parent is None if the parent in the cbf file
        is ".".  Parent is the axis that the top level axis in this chain of dependent axis depends on
        """

        cbf = self._get_cbf_handle()
        cob = self._get_change_of_basis(axis_id)

        parent_id = cbf.get_axis_depends_on(axis_id)

        if parent_id == b".":
            return None, cob

        eq_comp = cbf.get_axis_equipment_component(axis_id)
        parent_eq_comp = cbf.get_axis_equipment_component(parent_id)
        if eq_comp == parent_eq_comp:
            non_matching_parent, parent_cob = self._get_cumulative_change_of_basis(
                parent_id
            )
            return non_matching_parent, parent_cob * cob

        return parent_id, cob

    def _add_panel_group(self, group_id, d):
        """Adds a panel group to the detector d.  If the group's parent hasn't been
        added yet, recursively add parents to the detector until the detector itself
        is reached.
        @param group_id name of a cbf axis
        @param d detector object
        """

        # group_id will only be "." if the panel being worked on has the same equipment_component name as the
        # last axis in the hierarchy, which isn't really sensible
        assert group_id != b"."

        name = group_id

        for subobj in d.iter_preorder():
            if subobj.get_name().encode() == name:
                return subobj

        parent, cob = self._get_cumulative_change_of_basis(group_id)

        if parent is None:
            pg = d.hierarchy()  # root object for the detector
            try:
                pg.get_D_matrix()  # test to see if we've initialized the detector basis yet
            except RuntimeError as e:
                assert "DXTBX_ASSERT(D_)" in str(e)
            else:
                assert (
                    False
                )  # shouldn't be reached.  Detector should be initialized only once.
        else:
            parent_pg = self._add_panel_group(parent, d)
            pg = parent_pg.add_group()

        # set up the dxtbx d matrix.  Note use of homogenous coordinates.
        origin = col((cob * col((0, 0, 0, 1)))[0:3])
        fast = col((cob * col((1, 0, 0, 1)))[0:3]) - origin
        slow = col((cob * col((0, 1, 0, 1)))[0:3]) - origin

        pg.set_local_frame(fast.elems, slow.elems, origin.elems)

        pg.set_name(name)
        return pg

    def _detector(self):
        """Return a working detector instance."""

        cbf = self._get_cbf_handle()

        d = Detector()

        # find the panel element names. Either array ids or section ids
        cbf.find_category(b"array_structure_list")
        try:
            cbf.find_column(b"array_section_id")
        except Exception as e:
            if "CBF_NOTFOUND" not in str(e):
                raise e
            cbf.find_column(b"array_id")

        panel_names = []
        for i in range(cbf.count_rows()):
            cbf.select_row(i)
            if cbf.get_typeofvalue() == b"null":
                continue

            val = cbf.get_value()
            if val not in panel_names:
                panel_names.append(val)

        # the cbf detector objects are not guaranteed to be in the same order
        # as this array of panel names. re-iterate, associating root axes of
        # detector objects with panel names
        detector_axes = []
        has_sections = cbf.has_sections()

        for i in range(len(panel_names)):
            cbf_detector = cbf.construct_detector(i)
            axis0 = cbf_detector.get_detector_surface_axes(0)
            detector_axes.append(axis0)
            cbf_detector.__swig_destroy__(cbf_detector)
        cbf.find_category(b"array_structure_list")
        array_ids_detectororder = []
        panel_names_detectororder = []
        for detector_axis in detector_axes:
            cbf.find_column(b"axis_set_id")
            cbf.find_row(detector_axis)
            if has_sections:
                try:
                    cbf.find_column(
                        b"array_id"
                    )  # mandatory, but not always actually there
                except Exception as e:
                    if "CBF_NOTFOUND" not in str(e):
                        raise
                    cbf.find_column(b"array_section")  # use as backup, non standard
                array_ids_detectororder.append(cbf.get_value())
                cbf.find_column(b"array_section_id")
            else:
                cbf.find_column(b"array_id")
            panel_names_detectororder.append(cbf.get_value())

        for panel_number, panel_name in enumerate(panel_names):
            cbf_detector = cbf.construct_detector(
                panel_names_detectororder.index(panel_name)
            )

            # code adapted below from dxtbx.model.detector.DetectorFactory.imgCIF_H
            pixel = (
                cbf_detector.get_inferred_pixel_size(1),
                cbf_detector.get_inferred_pixel_size(2),
            )

            axis0 = cbf_detector.get_detector_surface_axes(0)
            axis1 = cbf_detector.get_detector_surface_axes(1)
            assert cbf.get_axis_depends_on(axis0) == axis1

            try:
                size = tuple(cbf.get_image_size_fs(i))
            except Exception as e:
                if "CBF_NOTFOUND" in str(e):
                    # no array data in the file, it's probably just a cbf header.  Get the image size elsewhere
                    size = [0, 0]
                    cbf.find_category(b"array_structure_list")
                    for axis in [axis0, axis1]:
                        cbf.find_column(b"axis_set_id")
                        cbf.find_row(axis)
                        cbf.find_column(b"precedence")
                        idx = int(cbf.get_value()) - 1
                        cbf.find_column(b"dimension")
                        size[idx] = int(cbf.get_value())
                    assert size[0] != 0 and size[1] != 0
                else:
                    raise e

            parent, cob = self._get_cumulative_change_of_basis(axis0)

            pg = self._add_panel_group(parent, d)

            p = pg.add_panel()

            fast = cbf.get_axis_vector(axis0)
            slow = cbf.get_axis_vector(axis1)
            origin = (cob * col((0, 0, 0, 1)))[0:3]

            p.set_local_frame(fast, slow, origin)

            try:
                max_trusted_value = cbf.get_overload(panel_number) - 1
                cbf.find_category(b"array_intensities")
                cbf.find_column(b"array_id")
                if has_sections:
                    cbf.find_row(array_ids_detectororder[panel_number])
                else:
                    cbf.find_row(panel_name)
                try:
                    cbf.find_column(b"underload")
                    offset = 0
                except Exception:
                    # By convention, if underload is not set, then assume the minimum
                    # trusted pixel is 1 more than the undefined pixel
                    cbf.find_column(b"undefined_value")
                    offset = 1
                min_trusted_value = cbf.get_doublevalue() + offset
                trusted_range = (min_trusted_value, max_trusted_value)
            except Exception as e:
                if "CBF_NOTFOUND" not in str(e):
                    raise
                trusted_range = (0.0, 1.0e6)

            try:
                cbf.find_column(b"gain")
                gain = cbf.get_doublevalue()
            except Exception as e:
                if "CBF_NOTFOUND" not in str(e):
                    raise
                gain = 1.0

            p.set_pixel_size(tuple(map(float, pixel)))
            p.set_image_size(size)
            p.set_trusted_range(tuple(map(float, trusted_range)))
            p.set_gain(gain)
            p.set_name(panel_name)
            # p.set_px_mm_strategy(px_mm) FIXME

            cbf_detector.__swig_destroy__(cbf_detector)
            del cbf_detector

        return d

    def _beam(self):
        """Return a working beam instance."""

        return self._beam_factory.imgCIF_H(self._get_cbf_handle())

    def get_raw_data(self):
        if self._raw_data is None:
            self._raw_data = []
            cbf = self._get_cbf_handle()
            cbf.find_category(b"array_structure")
            cbf.find_column(b"encoding_type")
            cbf.select_row(0)
            types = []
            for i in range(cbf.count_rows()):
                types.append(cbf.get_value())
                cbf.next_row()
            assert len(types) == cbf.count_rows()

            # read the data
            data = {}
            cbf.find_category(b"array_data")
            for i in range(cbf.count_rows()):
                cbf.find_column(b"array_id")
                name = cbf.get_value()

                cbf.find_column(b"data")
                assert cbf.get_typeofvalue().find(b"bnry") > -1

                if types[i] == b"signed 32-bit integer":
                    array_string = cbf.get_integerarray_as_string()
                    nelem = int(len(array_string) / 4)
                    array = flex.int(struct.unpack("%di" % nelem, array_string))
                    parameters = cbf.get_integerarrayparameters_wdims_fs()
                    array_size = (parameters[11], parameters[10], parameters[9])
                elif types[i] == b"signed 64-bit real IEEE":
                    array_string = cbf.get_realarray_as_string()
                    nelem = int(len(array_string) / 8)
                    array = flex.double(struct.unpack("%dd" % nelem, array_string))
                    parameters = cbf.get_realarrayparameters_wdims_fs()
                    array_size = (parameters[7], parameters[6], parameters[5])
                else:
                    return None  # type not supported

                array.reshape(flex.grid(*array_size))
                data[name] = array
                cbf.next_row()

            # extract the data for each panel
            if cbf.has_sections():
                section_shapes = {}
                for i in range(cbf.count_rows()):
                    cbf.find_column(b"id")
                    section_name = cbf.get_value()
                    if section_name not in section_shapes:
                        section_shapes[section_name] = {}
                    cbf.find_column(b"array_id")
                    if "array_id" not in section_shapes[section_name]:
                        section_shapes[section_name]["array_id"] = cbf.get_value()
                    else:
                        assert (
                            section_shapes[section_name]["array_id"] == cbf.get_value()
                        )
                    cbf.find_column(b"index")
                    axis_index = int(cbf.get_value()) - 1
                    cbf.find_column(b"start")
                    axis_start = int(cbf.get_value()) - 1
                    cbf.find_column(b"end")
                    axis_end = int(cbf.get_value())

                    section_shapes[section_name][axis_index] = slice(
                        axis_start, axis_end
                    )
                    cbf.next_row()

                for section_name in section_shapes:
                    section_shape = section_shapes[section_name]
                    section = data[section_shape["array_id"]][
                        section_shape[2], section_shape[1], section_shape[0]
                    ]
                    section.reshape(flex.grid(section.focus()[-2], section.focus()[-1]))
                    self._raw_data.append(section)
            else:
                for key in data:
                    data[key].reshape(
                        flex.grid(data[key].focus()[-2], data[key].focus()[-1])
                    )
                    self._raw_data.append(data[key])

            d = self.get_detector()
            assert len(d) == len(self._raw_data)

        return tuple(self._raw_data)


class FormatCBFMultiTileHierarchyStill(
    FormatCBFMultiTileStill, FormatCBFMultiTileHierarchy
):
    """An image reading class for full CBF format images i.e. those from
    a variety of cameras which support this format. Custom derived from
    the FormatStill to handle images without a gonimeter or scan"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBFMultiTileHierarchy.get_cbf_header(image_file)

        # According to ImageCIF, "Data items in the DIFFRN_MEASUREMENT_AXIS
        # category associate axes with goniometers."
        # http://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Cdiffrn_measurement_axis.html
        if "diffrn_measurement_axis" in header:
            return False
        return True


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMultiTileHierarchy.understand(arg))
